#CVS: $Id: DCverPrettyCode.pm,v 1.4 2009/03/31 08:10:58 depaulfm Exp $

# LICENSE AGREEMENT
#
#  FileName                   [DCverPrettyCode.pm]
#
#  PackageName                [BackSpace]
#
#  Synopsis                   [Parses Verilog code generated by Synopsys Design ]
#                             [Compiler and 'clean up' extra- spaces, lines;    ]
#                             [Only needed for BackSpace pre-processing         ]
#
#  Author                     [FLAVIO M DE PAULA]
#
#  Copyright                  [Copyright (C) 2009 University of British Columbia]
# 
#  This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

package DCverPrettyCode;

use 5.008;
use strict;
use warnings;
use Getopt::Long;
use Pod::Usage;
use File::Basename;
use Switch;

use lib "$ENV{'BS_PERL_MODULES'}";

use MyFileHandlers;
use ExtractVars;


###
# Private data to this class 
###
use vars qw (
             );

BEGIN {
    use Exporter   ();
    use vars       qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);
    
    $VERSION = '0.01';
    
    @ISA         = qw(Exporter);
    @EXPORT      = qw(&cleanUpCode); #export functions e.g. qw(&func1 &func2 &func4);
    %EXPORT_TAGS = ( );                            # eg: TAG => [ qw!name1 name2! ],
    
    # your exported package globals go here,
    # as well as any optionally exported functions
    @EXPORT_OK   =  qw(); #qw($Var1 %Hashit &func3);
}
use vars      @EXPORT_OK;

######################################
# Class constructor
#
######################################
sub new {
  my $package = shift;
  return bless({}, $package);
}



######################################## 
# SUBROUTINES
########################################
#
#
#
sub cleanUpCode{
    my $self = shift;
    my $ptr = shift;
    my 	@file_bare_bones = @$ptr;

    my $i;
    my $j;
    
    my $new_line = undef;
    my @cleaned_file = ();

    my $answ = 'n';
    my $outfile = ".work/cleaned.v";
    my $FHO;
    my $overwrite = 1;

#    if(-e $outfile) {
#	print "WARNING: Do you want to overwrite $outfile (y/n)?\n";
#	read STDIN, $answ, 2;
#	if($answ =~ /^y/i){ 
#	    $overwrite = 1;
#	}else{
#	    $overwrite = 0;
#	}
#    }else{
#	$overwrite = 1;
#    }
    
    if($overwrite) { #either file does not exist or want to overwrite
	print "... Generating pretty-verilog code\n";
	#
	# flatten wires to single-bits
	for($j = 0; $j <= $#file_bare_bones; $j = $j +1){
	    if($file_bare_bones[$j] =~ /wire\s+\[([0-9]+)\:([0-9]+)\]\s+([a-zA-Z0-9_]+)/){
		my $msb = $1;
		my $lsb = $2;
		my $net = $3;
		my $tmp = '';
		for($i = $lsb; $i < $msb; $i = $i +1) {
		    $tmp .=  $net . "_" . $i . "_, ";
		}
		$tmp .=  $net . "_" . $i . "_;\n";
		$file_bare_bones[$j] = "wire " . $tmp;
	    }
	}
	#
	# Replace \ and / from net names with _            #### This is new
	for($i = 0; $i <= $#file_bare_bones; $i = $i +1){  #### not removing
	    while ($file_bare_bones[$i] =~ /[\\\/]/){  #### [ and ]
		$file_bare_bones[$i] =~ s/[\\\/]/\_/;  #### because H_* are
		#print "$file_bare_bones[$i]\n";       #### arrays
	    }
	}
	#
	# Make instantiations one-single line instead of two
	$i = 0;
	$j=1;
	while($i <= $#file_bare_bones) {
	    #print "cleanup1: $file_bare_bones[$i]\n";
	    if($file_bare_bones[$i] =~  /\s*$ExtractVars::CELL/){
		$file_bare_bones[$i] =~ s/([\s+\t]+)$//;
		if($file_bare_bones[$i] !~ /\;$/) {
		    $file_bare_bones[$i] =~ s/([\s+\t]+)$//;
		    $j =$i+1;
		    $new_line = $file_bare_bones[$i] . $file_bare_bones[$j];
		    while(($new_line !~ /;/) && ($file_bare_bones[$j] !~ /endmodule/)){
			$new_line =~ s/([\s+\t]+)$//;
			#if ($j lt ($#file_bare_bones - 1)) { $j = $j + 1;}
			$j = $j + 1;
			$new_line =  $new_line . $file_bare_bones[$j];
			#print "j=$j***$new_line***\n";
		    }
		    if($new_line !~ /\n/) {$new_line .=  "\n"; }
		    $i = $j+1;	
		    push @cleaned_file, $new_line;
		    #print "cleanup2:$i:$file_bare_bones[$i]\n";
		    #print "cleanup3:$i:$new_line\n";
		    #print "cleanup1:$i\n";
		    #print "$new_line";
		    $new_line = undef;
		}else{
		    push @cleaned_file, $file_bare_bones[$i] . "\n";
		    $i = $i + 1;
		}
	    }else{    
		push @cleaned_file, $file_bare_bones[$i];
		$i = $i+1;
	    }
	}
	#
	# Dump cleaned.v into .work directory
	#
	open($FHO,">$outfile") or die "Cannot write $outfile";
	for($i = 0; $i <= $#cleaned_file; $i = $i +1){
	    print $FHO "$cleaned_file[$i]";
	}
	close $FHO;
    }else{##else don't overwrite cleaned.v
	#
	# Read cleaned,v file from.work directory
	#
	my $obj = new MyFileHandlers;my $currdir = $obj->MFHcurdir; my $workdir = $obj->MFHconcatdir2($currdir,".work");
	$obj->MFHfind($workdir,"cleaned.v");
	my @files = $obj->MFHlistoffiles;
	$obj->MFHreadfile($files[0]);
	@cleaned_file = $obj->MFHfile;
    }
    return @cleaned_file;
}

1;
