#CVS: $Id: ExtractVars.pm,v 1.4 2009/05/12 01:03:42 depaulfm Exp $

# LICENSE AGREEMENT
#
#  FileName                   [ExtractVars.pm]
#
#  PackageName                [BackSpace]
#
#  Synopsis                   [Extract combinational logic and latches from     ]
#                             [Synopsys Design Compiler formatted verilog       ]
#
#  Author                     [FLAVIO M DE PAULA]
#
#  Copyright                  [Copyright (C) 2009 University of British Columbia]
# 
#  This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

package ExtractVars;

use 5.008;
use strict;
use warnings;
use File::Find;
use File::Spec;


###
# Private data to this class 
###
use vars qw (   $AND2
		$AND3
		$AND4
		$NAND2
		$NAND3
		$NAND4
		$NAND5
		$OR2
		$OR3
		$OR4
		$NOR2
		$NOR3
		$NOR4
		$NOR5
		$NOT
		$DFF1
		$DFF2
		$DFF4
		$PORTLIST
                $BIT
		$INSTANCE_NAME	
                $SIGNAL_NAME
                $SIGNAL_NAME2
		);

BEGIN {
    use Exporter   ();
    use vars       qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);
    
    $VERSION = '0.01';
    
    @ISA         = qw(Exporter);
    @EXPORT      = qw(&getLatchVars &getCombVars &defineReset); #export functions e.g. qw(&func1 &func2 &func4);
    %EXPORT_TAGS = ( );                            # eg: TAG => [ qw!name1 name2! ],
    
    # your exported package globals go here,
    # as well as any optionally exported functions
    @EXPORT_OK   =  qw($CELL $ALLDFF $RESET); #qw($Var1 %Hashit &func3);
}
use vars      @EXPORT_OK;

# initalize package globals, first exported ones
#$Var1   = '';
#%Hashit = ();

# then the others (which are still accessible as $Some::Module::stuff)
#$stuff  = '';
#@more   = ();

# all file-scoped lexicals must be created before
# the functions below that use them.


######################################
# Class constructor
#
######################################
sub new {
  my $package = shift;
  return bless({}, $package);
}

#
# Dictionary 
#
# private 
$AND2      = '\bAN2isc\b';
$AND3      = '\bAN3isc\b';
$AND4      = '\bAN4isc\b';
$NAND2     = '\bND2isc\b';
$NAND3     = '\bND3isc\b';
$NAND4     = '\bND4isc\b';
$NAND5     = '\bND5isc\b';
$OR2       = '\bOR2isc\b';
$OR3       = '\bOR3isc\b';
$OR4       = '\bOR4isc\b';
$NOR2      = '\bNR2isc\b';
$NOR3      = '\bNR3isc\b';
$NOR4      = '\bNR4isc\b';
$NOR5      = '\bNR5isc\b';
$NOT       = '\bIVisc\b';
$DFF1      = '\bFD1isc\b';
$DFF2      = '\bFD2isc\b';
$DFF4      = '\bFD4isc\b';
$PORTLIST =  '\(.*\)';
#$INSTANCE_NAME = '\b[a-zA-Z0-9_]+\b';
$BIT      = '\[[0-9]+\]';
#$INSTANCE_NAME = "[a-zA-Z0-9_]+.($BIT){0,2}";
#$SIGNAL_NAME   = "[a-zA-Z0-9_]+.$BIT{0,2}?";

$INSTANCE_NAME = "[a-zA-Z0-9_]+(?:$BIT){0,2}";
$SIGNAL_NAME   = "[a-zA-Z0-9_]+(?:$BIT){0,2}";
$SIGNAL_NAME2   = "[a-zA-Z0-9_\']+(?:$BIT){0,2}";

$RESET     = "UNDEFINED";
# public  
$CELL      = 'AN2isc|AN3isc|AN4isc|ND2isc|ND3isc|ND4isc|ND5isc|OR2isc|OR3isc|OR4isc|NR2isc|NR3isc|NR4isc|NR5isc|IVisc|FD1isc|FD2isc|FD4isc';
$ALLDFF    = 'FD1isc|FD2isc|FD4isc';    


##############
# PUBLIC SUBROUTINES 
##############

#########################################
# Combinational Logic
#########################################
#
#
#
sub getCombVars{
    my $self = shift;
    my $line = shift;

    my $list = undef;
    my $i;
    my @final_list;

    $list = processDCVerComb($line);
    if(!defined($list)) { die "ERROR: $line did not generate any list\n";}
    $list =~ s/\s+//g;
    
    my @splitter = split(',',$list);
    for($i=0;$i<=$#splitter; $i=$i+1){
	push @final_list, $splitter[$i];
    }
    return @final_list;

}
#########################################
# Latches
#########################################
#
#
#
sub getLatchVars{
    my $self = shift;

    my $line = shift;

    my $list = undef;
    my $i;
    my @final_list = ();

    $list = processDCVerLatches($line);
    if(!defined($list)) { die "ERROR: $line did not generate any list\n";}
    $list =~ s/\s+//g;
    
    
    if($list !~ ''){
	my @splitter = split(',',$list);
#	print " what is in $splitter[$#splitter]\n";
	# when parsing FDx I switched the position of .D so that is the last element
	#   push @final_list, $splitter[$#splitter];
	for($i=2;$i<=$#splitter; $i=$i+1){
	    push @final_list, $splitter[$i];
	}
    }

    return @final_list;
}

##################################################################################################
# (PRIVATE) SUBROUTINES FOR DESIGN COMPILER'S VERILOG
##################################################################################################
#
#
#
sub processDCVerComb{
    my $line = shift;

    my $i = 0;
    my $port_list;
    my $outline = undef;
    my $inst_name;
    #print "$line";
    if ($line =~ /$AND2\s+($INSTANCE_NAME)\s*($PORTLIST);/){	
	$inst_name = $1;
	$port_list = $2;
	
	$outline = parseAND2($i,$port_list);
	if(!defined($outline)) {die "AND2 parse problem\n";}
	#push @new_file,   $AND2 . " " . $inst_name . "$outline" . ";\n";
    }else{
	if ($line =~ /$AND3\s+($INSTANCE_NAME)\s*($PORTLIST);/){	
	    $inst_name = $1;
	    $port_list = $2;
	    #print "$inst_name -- $port_list\n";
	    $outline = parseAND3($i,$port_list);
	    if(!defined($outline)) {die "AND3 parse problem\n";}
	    #push @new_file,   $AND3 . " " . $inst_name . "$outline" . ";\n";
	}else{
	    if ($line =~ /$AND4\s+($INSTANCE_NAME)\s*($PORTLIST);/){	
		$inst_name = $1;
		$port_list = $2;
		$outline = parseAND4($i,$port_list);
	    if(!defined($outline)) {die "AND4 parse problem\n";}
		#push @new_file,   $AND4 . " " . $inst_name . "$outline" . ";\n";
	    }else{
		if ($line =~ /$NAND2\s+($INSTANCE_NAME)\s*($PORTLIST);/){	
		    $inst_name = $1;
		    $port_list = $2;
		    $outline = parseNAND2($i,$port_list);
		    if(!defined($outline)) {die "ND2 parse problem\n";}
		    #push @new_file,   $NAND2 . " " . $inst_name . "$outline" . ";\n";
		}else{
		    if ($line =~ /$NAND3\s+($INSTANCE_NAME)\s*($PORTLIST);/){	
			$inst_name = $1;
			$port_list = $2;
			$outline = parseNAND3($i,$port_list);
			if(!defined($outline)) {die "ND3 parse problem\n";}
			#push @new_file,   $NAND3 . " " . $inst_name . "$outline" . ";\n";
		    }else{
			if ($line =~ /$NAND4\s+($INSTANCE_NAME)\s*($PORTLIST);/){	
			    $inst_name = $1;
			    $port_list = $2;
			    $outline = parseNAND4($i,$port_list);
			    if(!defined($outline)) {die "ND4 parse problem\n";}
			    #push @new_file,   $NAND4 . " " . $inst_name . "$outline" . ";\n";
			}else{
			    if ($line =~ /$NAND5\s+($INSTANCE_NAME)\s*($PORTLIST);/){	
				$inst_name = $1;
				$port_list = $2;
				$outline = parseNAND5($i,$port_list);
				if(!defined($outline)) {die "ND5 parse problem\n";}
				#push @new_file,   $NAND5 . " " . $inst_name . "$outline" . ";\n";
			    }else{
				if ($line =~ /$OR2\s+($INSTANCE_NAME)\s*($PORTLIST);/){	
				    $inst_name = $1;
				    $port_list = $2;
				    $outline = parseOR2($i,$port_list);
				    if(!defined($outline)) {die "OR2 parse problem\n";}
				    #push @new_file,   $OR2 . " " . $inst_name . "$outline" . ";\n";
				}else{
				    if ($line =~ /$OR3\s+($INSTANCE_NAME)\s*($PORTLIST);/){	
					$inst_name = $1;
					$port_list = $2;
					$outline = parseOR3($i,$port_list);
					if(!defined($outline)) {die "OR3 parse problem\n";}
					#push @new_file,   $OR3 . " " . $inst_name . "$outline" . ";\n";
				    }else{
					if ($line =~ /$OR4\s+($INSTANCE_NAME)\s*($PORTLIST);/){	
					    $inst_name = $1;
					    $port_list = $2;
					    $outline = parseOR4($i,$port_list);
					    if(!defined($outline)) {die "OR4 parse problem\n";}
					    #push @new_file,   $OR4 . " " . $inst_name . "$outline" . ";\n";
					}else{
					    if ($line =~ /$NOR2\s+($INSTANCE_NAME)\s*($PORTLIST);/){	
						$inst_name = $1;
						$port_list = $2;
						$outline = parseNOR2($i,$port_list);
						if(!defined($outline)) {die "ND3 parse problem\n";}
						#push @new_file,   $NOR2 . " " . $inst_name . "$outline" . ";\n";
					    }else{
						if ($line =~ /$NOR3\s+($INSTANCE_NAME)\s*($PORTLIST);/){	
						    $inst_name = $1;
						    $port_list = $2;
						    $outline = parseNOR3($i,$port_list);
						    if(!defined($outline)) {die "NOR3 parse problem\n";}
						    #push @new_file,   $NOR3 . " " . $inst_name . "$outline" . ";\n";
						}else{
						    if ($line =~ /$NOR4\s+($INSTANCE_NAME)\s*($PORTLIST);/){	
							$inst_name = $1;
							$port_list = $2;
							$outline = parseNOR4($i,$port_list);
							if(!defined($outline)) {die "NOR4 parse problem\n";}
							#push @new_file,   $NOR4 . " " . $inst_name . "$outline" . ";\n";
						    }else{
							if ($line =~ /$NOR5\s+($INSTANCE_NAME)\s*($PORTLIST);/){	
							    $inst_name = $1;
							    $port_list = $2;
							    $outline = parseNOR5($i,$port_list);
							    if(!defined($outline)) {die "NOR5 parse problem\n";}
							    #push @new_file,   $NOR5 . " " . $inst_name . "$outline" . ";\n";
							}else{
							    if ($line =~ /$NOT\s+($INSTANCE_NAME)\s*($PORTLIST);/){	
								$inst_name = $1;
								$port_list = $2;
								$outline = parseNOT($i,$port_list);
								if(!defined($outline)) {die "NOT parse problem\n";}
								#push @new_file,   $NOT . " " . $inst_name . "$outline" . ";\n";
							    }
							}
						    }
						}
					    }
					}
				    }
				}
			    }
			}
		    }
		}
	    }
	}
    }
    #print $outline . "\n";
    return $outline;
}

#
#
#
sub parseAND2{
    my ($inst, $line) = @_;
    my $inp1;
    my $inp2;
    my $out;
    my $ports;
    if($line =~ /\.A\(\s*($SIGNAL_NAME2\s*)\),\s+\.B\(\s*($SIGNAL_NAME2\s*)\),\s+\.Z\(\s*($SIGNAL_NAME\s*)\s*\)/) {
	$inp1 = $1;
	$inp2 = $2;
	$out  = $3;

	if($inp1 =~ /1\'b0/){$inp1 = "GND";}
	else {if($inp1 =~ /1\'b1/){ $inp1 = "VCC";}}

	if($inp2 =~ /1\'b0/){ $inp2 = "GND";}
	else {if($inp2 =~ /1\'b1/){ $inp2 = "VCC";}}

	$ports = $inp1 . ", " . $inp2 . ", " . $out;

    }else{ 
	die "Missing port in AN2 instance at line $line\n";
    }
    return $ports;
}
#
#
#
sub parseAND3{
    my ($inst, $line) = @_;
    my $inp1;
    my $inp2;
    my $inp3;
    my $out;
    my $ports;
    if($line =~ /\.A\(\s*($SIGNAL_NAME2\s*)\),\s+\.B\(\s*($SIGNAL_NAME2\s*)\),\s+\.C\(\s*($SIGNAL_NAME2\s*)\),\s+\.Z\(\s*($SIGNAL_NAME\s*)\s*\)/) {
	$inp1 = $1;
	$inp2 = $2;
	$inp3 = $3;
	$out  = $4;

	if($inp1 =~ /1\'b0/){$inp1 = "GND";}
	else {if($inp1 =~ /1\'b1/){ $inp1 = "VCC";}}

	if($inp2 =~ /1\'b0/){ $inp2 = "GND";}
	else {if($inp2 =~ /1\'b1/){ $inp2 = "VCC";}}

	if($inp3 =~ /1\'b0/){ $inp3 = "GND";}
	else {if($inp3 =~ /1\'b1/){ $inp3 = "VCC";}}

	$ports = $inp1 . ", " . $inp2 . ", " . $inp3 . ", " . $out;
    }else{ 
	die "Missing port in AN3 instance at line $line\n";
    }
    
    return $ports;
}
#
#
#
sub parseAND4{
    my ($inst, $line) = @_;
    my $inp1;
    my $inp2;
    my $inp3;
    my $inp4;
    my $out;
    my $ports;
    if($line =~ /\.A\(\s*($SIGNAL_NAME2\s*)\),\s+\.B\(\s*($SIGNAL_NAME2\s*)\),\s+\.C\(\s*($SIGNAL_NAME2\s*)\),\s+\.D\(\s*($SIGNAL_NAME2\s*)\),\s+\.Z\(\s*($SIGNAL_NAME\s*)\s*\)/) {
	$inp1 = $1;
	$inp2 = $2;
	$inp3 = $3;
	$inp4 = $4;
	$out  = $5;

	if($inp1 =~ /1\'b0/){$inp1 = "GND";}
	else {if($inp1 =~ /1\'b1/){ $inp1 = "VCC";}}

	if($inp2 =~ /1\'b0/){ $inp2 = "GND";}
	else {if($inp2 =~ /1\'b1/){ $inp2 = "VCC";}}

	if($inp3 =~ /1\'b0/){ $inp3 = "GND";}
	else {if($inp3 =~ /1\'b1/){ $inp3 = "VCC";}}

	if($inp4 =~ /1\'b0/){ $inp4 = "GND";}
	else {if($inp4 =~ /1\'b1/){ $inp4 = "VCC";}}

	$ports = $inp1 . ", " . $inp2 . ", " . $inp3 . ", " . $inp4 . ", " . $out;
    }else{ 
	die "Missing port in AN4 instance at line $line\n";
    }
    
    return $ports;

}
#
#
#
sub parseNAND2{
    my ($inst, $line) = @_;
    my $inp1;
    my $inp2;
    my $out;
    my $ports;
    if($line =~ /\.A\(\s*($SIGNAL_NAME2\s*)\),\s+\.B\(\s*($SIGNAL_NAME2\s*)\),\s+\.Z\(\s*($SIGNAL_NAME)\s*\)/) {
	$inp1 = $1;
	$inp2 = $2;
	$out  = $3;

	if($inp1 =~ /1\'b0/){$inp1 = "GND";}
	else {if($inp1 =~ /1\'b1/){ $inp1 = "VCC";}}

	if($inp2 =~ /1\'b0/){ $inp2 = "GND";}
	else {if($inp2 =~ /1\'b1/){ $inp2 = "VCC";}}

	$ports = $inp1 . ", " . $inp2 . ", " . $out;
    }else{ 
	die "Missing port in ND2 instance at line $line\n";
    }
    
    return $ports;
}
#
#
#
sub parseNAND3{
    my ($inst, $line) = @_;
    my $inp1;
    my $inp2;
    my $inp3;
    my $out;
    my $ports;
    if($line =~ /\.A\(\s*($SIGNAL_NAME2\s*)\),\s+\.B\(\s*($SIGNAL_NAME2\s*)\),\s+\.C\(\s*($SIGNAL_NAME2\s*)\),\s+\.Z\(\s*($SIGNAL_NAME\s*)\s*\)/) {
	$inp1 = $1;
	$inp2 = $2;
	$inp3 = $3;
	$out  = $4;

	if($inp1 =~ /1\'b0/){$inp1 = "GND";}
	else {if($inp1 =~ /1\'b1/){ $inp1 = "VCC";}}

	if($inp2 =~ /1\'b0/){ $inp2 = "GND";}
	else {if($inp2 =~ /1\'b1/){ $inp2 = "VCC";}}

	if($inp3 =~ /1\'b0/){ $inp3 = "GND";}
	else {if($inp3 =~ /1\'b1/){ $inp3 = "VCC";}}

 	$ports = $inp1 . ", " . $inp2 . ", " . $inp3 . ", " . $out;
    }else{ 
	die "Missing port in ND3 instance at line $line\n";
    }
    
    return $ports;
}
#
#
#
sub parseNAND4{
    my ($inst, $line) = @_;
    my $inp1;
    my $inp2;
    my $inp3;
    my $inp4;
    my $out;
    my $ports;
    if($line =~ /\.A\(\s*($SIGNAL_NAME2\s*)\),\s+\.B\(\s*($SIGNAL_NAME2\s*)\),\s+\.C\(\s*($SIGNAL_NAME2\s*)\),\s+\.D\(\s*($SIGNAL_NAME2\s*)\),\s+\.Z\(\s*($SIGNAL_NAME\s*)\s*\)/) {
	$inp1 = $1;
	$inp2 = $2;
	$inp3 = $3;
	$inp4 = $4;
	$out  = $5;
	if($inp1 =~ /1\'b0/){$inp1 = "GND";}
	else {if($inp1 =~ /1\'b1/){ $inp1 = "VCC";}}

	if($inp2 =~ /1\'b0/){ $inp2 = "GND";}
	else {if($inp2 =~ /1\'b1/){ $inp2 = "VCC";}}

	if($inp3 =~ /1\'b0/){ $inp3 = "GND";}
	else {if($inp3 =~ /1\'b1/){ $inp3 = "VCC";}}

	if($inp4 =~ /1\'b0/){ $inp4 = "GND";}
	else {if($inp4 =~ /1\'b1/){ $inp4 = "VCC";}}


 	$ports = $inp1 . ", " . $inp2 . ", " . $inp3 . ", " . $inp4 . ", " . $out;
    }else{ 
	die "Missing port in ND4 instance at line $line\n";
    }
    
    return $ports;

}
#
#
#
sub parseNAND5{
    my ($inst, $line) = @_;
    my $inp1;
    my $inp2;
    my $inp3;
    my $inp4;
    my $inp5;
    my $out;
    my $ports;
    if($line =~ /\.A\(\s*($SIGNAL_NAME2\s*)\),\s+\.B\(\s*($SIGNAL_NAME2\s*)\),\s+\.C\(\s*($SIGNAL_NAME2\s*)\),\s+\.D\(\s*($SIGNAL_NAME2\s*)\),\s+\.E\(\s*($SIGNAL_NAME2\s*)\),\s+\.Z\(\s*($SIGNAL_NAME\s*)\s*\)/) {
	$inp1 = $1;
	$inp2 = $2;
	$inp3 = $3;
	$inp4 = $4;
	$inp5 = $5;
	$out  = $6;

	if($inp1 =~ /1\'b0/){$inp1 = "GND";}
	else {if($inp1 =~ /1\'b1/){ $inp1 = "VCC";}}

	if($inp2 =~ /1\'b0/){ $inp2 = "GND";}
	else {if($inp2 =~ /1\'b1/){ $inp2 = "VCC";}}

	if($inp3 =~ /1\'b0/){ $inp3 = "GND";}
	else {if($inp3 =~ /1\'b1/){ $inp3 = "VCC";}}

	if($inp4 =~ /1\'b0/){ $inp4 = "GND";}
	else {if($inp4 =~ /1\'b1/){ $inp4 = "VCC";}}

	if($inp5 =~ /1\'b0/){ $inp5 = "GND";}
	else {if($inp5 =~ /1\'b1/){ $inp5 = "VCC";}}

 	$ports = $inp1 . ", " . $inp2 . ", " . $inp3 . ", " . $inp4 . ", " . $inp5 . ", " . $out;

    }else{ 
	die "Missing port in ND5 instance at line $line\n";
    }
    
    return $ports;
}
#
#
#
sub parseOR2{
    my ($inst, $line) = @_;
    my $inp1;
    my $inp2;
    my $out;
    my $ports;
    if($line =~ /\.A\(\s*($SIGNAL_NAME2\s*)\),\s+\.B\(\s*($SIGNAL_NAME2\s*)\),\s+\.Z\(\s*($SIGNAL_NAME\s*)\s*\)/) {
	$inp1 = $1;
	$inp2 = $2;
	$out  = $3;

	if($inp1 =~ /1\'b0/){$inp1 = "GND";}
	else {if($inp1 =~ /1\'b1/){ $inp1 = "VCC";}}

	if($inp2 =~ /1\'b0/){ $inp2 = "GND";}
	else {if($inp2 =~ /1\'b1/){ $inp2 = "VCC";}}

 	$ports = $inp1 . ", " . $inp2 . ", " . $out;
    }else{ 
	die "Missing port in OR2 instance at line $line\n";
    }
    
    return $ports;
}
#
#
#
sub parseOR3{
    my ($inst, $line) = @_;
    my $inp1;
    my $inp2;
    my $inp3;
    my $out;
    my $ports;
   if($line =~ /\.A\(\s*($SIGNAL_NAME2\s*)\),\s+\.B\(\s*($SIGNAL_NAME2\s*)\),\s+\.C\(\s*($SIGNAL_NAME2\s*)\),\s+\.Z\(\s*($SIGNAL_NAME\s*)\s*\)/) {
	$inp1 = $1;
	$inp2 = $2;
	$inp3 = $3;
	$out  = $4;
	if($inp1 =~ /1\'b0/){$inp1 = "GND";}
	else {if($inp1 =~ /1\'b1/){ $inp1 = "VCC";}}

	if($inp2 =~ /1\'b0/){ $inp2 = "GND";}
	else {if($inp2 =~ /1\'b1/){ $inp2 = "VCC";}}

	if($inp3 =~ /1\'b0/){ $inp3 = "GND";}
	else {if($inp3 =~ /1\'b1/){ $inp3 = "VCC";}}

 	$ports = $inp1 . ", " . $inp2 . ", " . $inp3 . ", " . $out;
    }else{ 
	die "Missing port in OR3 instance at line $line\n";
    }
    
    return $ports;
}
#
#
#
sub parseOR4{
    my ($inst, $line) = @_;
    my $inp1;
    my $inp2;
    my $inp3;
    my $inp4;
    my $out;
    my $ports;
    if($line =~ /\.A\(\s*($SIGNAL_NAME2\s*)\),\s+\.B\(\s*($SIGNAL_NAME2\s*)\),\s+\.C\(\s*($SIGNAL_NAME2\s*)\),\s+\.D\(\s*($SIGNAL_NAME2\s*)\),\s+\.Z\(\s*($SIGNAL_NAME\s*)\s*\)/) {
	$inp1 = $1;
	$inp2 = $2;
	$inp3 = $3;
	$inp4 = $4;
	$out  = $5;
	if($inp1 =~ /1\'b0/){$inp1 = "GND";}
	else {if($inp1 =~ /1\'b1/){ $inp1 = "VCC";}}

	if($inp2 =~ /1\'b0/){ $inp2 = "GND";}
	else {if($inp2 =~ /1\'b1/){ $inp2 = "VCC";}}

	if($inp3 =~ /1\'b0/){ $inp3 = "GND";}
	else {if($inp3 =~ /1\'b1/){ $inp3 = "VCC";}}

	if($inp4 =~ /1\'b0/){ $inp4 = "GND";}
	else {if($inp4 =~ /1\'b1/){ $inp4 = "VCC";}}

 	$ports = $inp1 . ", " . $inp2 . ", " . $inp3 . ", " . $inp4 . ", " . $out;
    }else{ 
	die "Missing port in OR4 instance at line $line\n";
    }
    
    return $ports;

}
#
#
#
sub parseNOR2{
    my ($inst, $line) = @_;
    my $inp1;
    my $inp2;
    my $out;
    my $ports;
    if($line =~ /\.A\(\s*($SIGNAL_NAME2\s*)\),\s+\.B\(\s*($SIGNAL_NAME2\s*)\),\s+\.Z\(\s*($SIGNAL_NAME\s*)\s*\)/) {
	$inp1 = $1;
	$inp2 = $2;
	$out  = $3;
	if($inp1 =~ /1\'b0/){$inp1 = "GND";}
	else {if($inp1 =~ /1\'b1/){ $inp1 = "VCC";}}

	if($inp2 =~ /1\'b0/){ $inp2 = "GND";}
	else {if($inp2 =~ /1\'b1/){ $inp2 = "VCC";}}


 	$ports = $inp1 . ", " . $inp2 . ", " . $out;
    }else{ 
	die "Missing port in NR2 instance at line $line\n";
    }
    return $ports;
}
#
#
#
sub parseNOR3{
    my ($inst, $line) = @_;
    my $inp1;
    my $inp2;
    my $inp3;
    my $out;
    my $ports;
    if($line =~ /\.A\(\s*($SIGNAL_NAME2\s*)\),\s+\.B\(\s*($SIGNAL_NAME2\s*)\),\s+\.C\(\s*($SIGNAL_NAME2\s*)\),\s+\.Z\(\s*($SIGNAL_NAME\s*)\s*\)/) {
	$inp1 = $1;
	$inp2 = $2;
	$inp3 = $3;
	$out  = $4;
	if($inp1 =~ /1\'b0/){$inp1 = "GND";}
	else {if($inp1 =~ /1\'b1/){ $inp1 = "VCC";}}

	if($inp2 =~ /1\'b0/){ $inp2 = "GND";}
	else {if($inp2 =~ /1\'b1/){ $inp2 = "VCC";}}

	if($inp3 =~ /1\'b0/){ $inp3 = "GND";}
	else {if($inp3 =~ /1\'b1/){ $inp3 = "VCC";}}

	$ports = $inp1 . ", " . $inp2 . ", " . $inp3 . ", " . $out;
    }else{ 
	die "Missing port in NR3 instance at line $line\n";
    }
    
    return $ports;
}
#
#
#
sub parseNOR4{
    my ($inst, $line) = @_;
    my $inp1;
    my $inp2;
    my $inp3;
    my $inp4;
    my $out;
    my $ports;
    if($line =~ /\.A\(\s*($SIGNAL_NAME2\s*)\),\s+\.B\(\s*($SIGNAL_NAME2\s*)\),\s+\.C\(\s*($SIGNAL_NAME2\s*)\),\s+\.D\(\s*($SIGNAL_NAME2\s*)\),\s+\.Z\(\s*($SIGNAL_NAME\s*)\s*\)/) {
       $inp1 = $1;
	$inp2 = $2;
	$inp3 = $3;
	$inp4 = $4;
	$out  = $5;
	if($inp1 =~ /1\'b0/){$inp1 = "GND";}
	else {if($inp1 =~ /1\'b1/){ $inp1 = "VCC";}}

	if($inp2 =~ /1\'b0/){ $inp2 = "GND";}
	else {if($inp2 =~ /1\'b1/){ $inp2 = "VCC";}}

	if($inp3 =~ /1\'b0/){ $inp3 = "GND";}
	else {if($inp3 =~ /1\'b1/){ $inp3 = "VCC";}}

	if($inp4 =~ /1\'b0/){ $inp4 = "GND";}
	else {if($inp4 =~ /1\'b1/){ $inp4 = "VCC";}}

 	$ports = $inp1 . ", " . $inp2 . ", " . $inp3 . ", " . $inp4 . ", " . $out;
    }else{ 
	die "Missing port in NR4 instance at line $line\n";
    }
    
    return $ports;
}
#
#
#
sub parseNOR5{
    my ($inst, $line) = @_;
    my $inp1;
    my $inp2;
    my $inp3;
    my $inp4;
    my $inp5;
    my $out;
    my $ports = '';
    if($line =~ /\.A\(\s*($SIGNAL_NAME2\s*)\),\s+\.B\(\s*($SIGNAL_NAME2\s*)\),\s+\.C\(\s*($SIGNAL_NAME2\s*)\),\s+\.D\(\s*($SIGNAL_NAME2\s*)\),\s+\.E\(\s*($SIGNAL_NAME2\s*)\),\s+\.Z\(\s*($SIGNAL_NAME\s*)\s*\)/) {
	$inp1 = $1;
	$inp2 = $2;
	$inp3 = $3;
	$inp4 = $4;
	$inp5 = $5;
	$out  = $6;
	if($inp1 =~ /1\'b0/){$inp1 = "GND";}
	else {if($inp1 =~ /1\'b1/){ $inp1 = "VCC";}}

	if($inp2 =~ /1\'b0/){ $inp2 = "GND";}
	else {if($inp2 =~ /1\'b1/){ $inp2 = "VCC";}}

	if($inp3 =~ /1\'b0/){ $inp3 = "GND";}
	else {if($inp3 =~ /1\'b1/){ $inp3 = "VCC";}}

	if($inp4 =~ /1\'b0/){ $inp4 = "GND";}
	else {if($inp4 =~ /1\'b1/){ $inp4 = "VCC";}}

	if($inp5 =~ /1\'b0/){ $inp5 = "GND";}
	else {if($inp5 =~ /1\'b1/){ $inp5 = "VCC";}}

 	$ports = $inp1 . ", " . $inp2 . ", " . $inp3 . ", " . $inp4 . ", " . $inp5 . ", " . $out;
    }else{ 
	die "Missing port in NR5 instance at line $line\n";
    }
    
    return $ports;
}
#
#
#
sub parseNOT{
    my ($inst, $line) = @_;
    my $inp1;
    my $out;
    my $ports;
    if($line =~ /\.A\(\s*($SIGNAL_NAME2\s*)\),\s+\.Z\(\s*($SIGNAL_NAME\s*)\s*\)/) {
	$inp1 = $1;
	$out  = $2;
	if($inp1 =~ /1\'b0/){$inp1 = "GND";}
	else {if($inp1 =~ /1\'b1/){ $inp1 = "VCC";}}

 	$ports = $inp1 . ", " . $out;

    }else{ 
	die "Missing port in IV instance at line $line\n";
    }
    
    return $ports;
}

#
#
#
sub parseDFF1{
    my ($inst, $line) = @_;
    my $inp1;
    my $inp2;
    my $out1;
    my $out2;
    my $ports;
    my $part1;
   if($line =~ /\.D\(\s*($SIGNAL_NAME\s*)\),\s+\.CP\(\s*($SIGNAL_NAME\s*)\),\s+\.Q\(\s*($SIGNAL_NAME\s*)\),\s+\.QN\(\s*($SIGNAL_NAME\s*)\)/) {
	$inp1 = $1;
	$inp2 = $2;
	$out1 = $3;
	$out2 = $4;
	$ports = $2 . ", .SD(DUMMY), " . $3 . ", " . $4 . ", " . $1;
    }else{ 
	if($line !~ /\.D\(\s*/)  { die "Missing port D in FD1 instance at line $line\n";}
	if($line !~ /\.CP\(\s*/) { die "Missing port CP in FD1 instance at line $line\n";}
	if($line !~ /\.Q\(/) {
	    if ($line =~ /\.D\(\s*($SIGNAL_NAME\s*)\),\s+\.CP\(\s*($SIGNAL_NAME\s*)\),\s+/){
		$part1 =  $2 . ", .SD(DUMMY), ";
		$inp1 = $1;
	    }
	    if ($line =~ /\.QN\(\s*($SIGNAL_NAME\s*).*\)/) {
		$ports = $part1 . $1 . ", " . $inp1;
	    }
	}
	if($line !~ /\.QN\(/) {
	 if ($line =~ /\.D\(\s*($SIGNAL_NAME\s*)\),\s+\.CP\(\s*($SIGNAL_NAME\s*)\),\s+\.Q\(\s*($SIGNAL_NAME\s*)\).*/){
	     $ports =  $2 . ", .SD(DUMMY), " . $3 . ", "  . $1;
	    }
	}
    }
   # print $ports . "\n";
    return $ports;
}

#
#
#
sub parseDFF2{
    my ($inst, $line) = @_;
    my $inp1;
    my $inp2;
    my $inp3;
    my $out1;
    my $out2;
    my $part1;
    my $ports = '';
    if($line =~ /\.D\(\s*($SIGNAL_NAME\s*)\),\s+\.CP\(\s*($SIGNAL_NAME\s*)\),\s+\.CD\(\s*($SIGNAL_NAME\s*)\),\s+\.Q\(\s*($SIGNAL_NAME\s*)\),\s+\.QN\(\s*($SIGNAL_NAME\s*)\)/) {
	$inp1 = $1;
	$inp2 = $2;
	$inp3 = $3;
	$out1 = $4;
	$out2 = $5;
	$ports =  $2 . ", " . $3 . ", " . $4 . ", " . $5 . ", " . $1;
    }else{ 
	if($line !~ /\.D\(\s*/)  { die "Missing port D in FD1 instance at line $line\n";}
	if($line !~ /\.CP\(\s*/) { die "Missing port CP in FD1 instance at line $line\n";}
	if($line !~ /\.CD\(/) { die "Missing port CP in FD1 instance at line $line\n";}
	if(($line !~ /\.Q\(/) && ($line !~ /\.QN\(/)){
	    if ($line =~ /\.D\(\s*($SIGNAL_NAME2\s*)\),\s+\.CP\(\s*($SIGNAL_NAME\s*)\),\s+\.CD\(\s*($SIGNAL_NAME\s*)\)\s*/){
		$ports =   $2 . ", " . $3 . ", " . $1;
	    }
	}else{
	    if($line !~ /\.Q\(/) {
		if ($line =~ /\.D\(\s*($SIGNAL_NAME2\s*)\),\s+\.CP\(\s*($SIGNAL_NAME\s*)\),\s+\.CD\(\s*($SIGNAL_NAME\s*)\),\s+/){
		    $part1 =   $2 . ", " . $3;
		    $inp1 = $1; 
		}
		if ($line =~ /\.QN\(\s*($SIGNAL_NAME\s*)\).*/) {
		    $ports = $part1 . ", " . $1 . ", " . $inp1;
		}
	    }
	    if($line !~ /\.QN\(/) {
		if ($line =~ /\.D\(\s*($SIGNAL_NAME2\s*)\),\s+\.CP\(\s*($SIGNAL_NAME\s*)\),\s+\.CD\(\s*($SIGNAL_NAME\s*)\),\s+\.Q\(\s*($SIGNAL_NAME\s*)\)\s*/){
		    $ports =  $2 . ", " . $3 . ", " . $4 .  ", " . $1;
		}
	    }
	}
    }
    #print $ports . "\n";
    return $ports;
}

#
#
#
sub parseDFF4{
    my ($inst, $line) = @_;

    my $inp1;
    my $inp2;
    my $inp3;
    my $out1;
    my $out2;
    my $part1;
    my $ports = '';
    if($line =~ /\.D\(\s*($SIGNAL_NAME\s*)\),\s+\.CP\(\s*($SIGNAL_NAME\s*)\),\s+\.SD\(\s*($SIGNAL_NAME\s*)\),\s+\.Q\(\s*($SIGNAL_NAME\s*)\),\s+\.QN\(\s*($SIGNAL_NAME\s*)\)/) {
	$inp1 = $1;
	$inp2 = $2;
	$inp3 = $3;
	$out1 = $4;
	$out2 = $5;
	$ports =  $2 . ", " . $3 . ", " . $4 . ", " . $5 . ", " . $1;
    }else{ 
	if($line !~ /\.D\(\s*/)  { die "Missing port D in FD1 instance at line $line\n";}
	if($line !~ /\.CP\(\s*/) { die "Missing port CP in FD1 instance at line $line\n";}
	if($line !~ /\.SD\(/) { die "Missing port CP in FD1 instance at line $line\n";}
	if(($line !~ /\.Q\(/) && ($line !~ /\.QN\(/)) {
	    if ($line =~ /\.D\(\s*($SIGNAL_NAME\s*)\),\s+\.CP\(\s*($SIGNAL_NAME\s*)\),\s+\.SD\(\s*($SIGNAL_NAME\s*)\)/){
		$ports =    $2 . ", " . $3 . ", " . $1;	
	    }
	}else{
	    if($line !~ /\.Q\(/) {
		if ($line =~ /\.D\(\s*($SIGNAL_NAME2\s*)\),\s+\.CP\(\s*($SIGNAL_NAME\s*)\),\s+\.SD\(\s*($SIGNAL_NAME\s*)\),\s+/){
		    $part1 =   $2 . ", " . $3;
		    $inp1  =  $1;
		}
		if ($line =~ /\.QN\(\s*($SIGNAL_NAME\s*)\).*/) {
		    $ports = $part1 . ", " . $1 . ", " . $inp1;
		}
	    }
	    if($line !~ /\.QN\(/) {
		if ($line =~ /\.D\(\s*($SIGNAL_NAME2\s*)\),\s+\.CP\(\s*($SIGNAL_NAME\s*)\),\s+\.SD\(\s*($SIGNAL_NAME\s*)\),\s+\.Q\(\s*($SIGNAL_NAME\s*)\)\s*/){
		    $ports =  $2 . ", " . $3 . ", " . $4 .  ", " . $1;
		}
	    }
	}
    }

    #print $ports . "\n";
    return $ports;
}
#
#
#
sub parseHARDCODE{
    my ($line) = shift;

    my $inp1;
    my $inp2;
    my $inp3;
    my $out1;
    my $out2;
    my $part1;
    my $ports = '';
    if($line =~ /\.D\(\s*(.*)\).*\.Q\((.*)\).*\.QN\((.*)\).*\)/) {
	$inp1 = $1;
	$inp2 = $2;
	$inp3 = $3;
	if($inp1 =~ /1\'/){
	    print "inp2=$inp2---inp3=$inp3---\n";
	    $ports =  "CP(DUMMY), " . "SD(DUMMY), " . $inp2 . ", " . $inp3 . ", " . "DONTCARE";
	}
    }else{
	if($line !~ /\.D\(\s*/)  { die "Missing port D in FD1 instance at line $line\n";}
	if($line !~ /\.CP\(\s*/) { die "Missing port CP in FD1 instance at line $line\n";}
	if($line !~ /\.(S|C)D\(/) {
	    if(($line !~ /\.Q\(/) && ($line !~ /\.QN\(/)) {
		if ($line =~ /\.D\(\s*(1\'.*)\),\s+\.CP\(\s*($SIGNAL_NAME\s*)\),/){
		    $inp1 = $1;
		    $ports =    $2 . ", " . "SD(DUMMY)" . ", " . "DONTCARE";	
		}
	    }else{
		if($line !~ /\.Q\(/) {
		    if ($line =~ /\.D\(\s*(1\'.*)\),\s+\.CP\(\s*($SIGNAL_NAME\s*)\),/){
			$part1 =   $2 . ", " . "SD(DUMMY)";
		    }
		    if ($line =~ /\.QN\(\s*($SIGNAL_NAME\s*)\).*\)/) {
			$ports = $part1 . ", " . $1 . ", " . "DONTCARE";
		    }
		}
		if($line !~ /\.QN\(/) {
		    if ($line =~ /\.D\(\s*(1\'.*)\),\s+\.CP\(\s*($SIGNAL_NAME\s*)\),\s+\.Q\(($SIGNAL_NAME\s*)\).*/){
			$ports =  $2 . ", " . "SD(DUMMY)" . ", " . $3 .  ", " . "DONTCARE";
		    }
		}
	    }
	}else{
	    if(($line !~ /\.Q\(/) && ($line !~ /\.QN\(/)) {
		if ($line =~ /\.D\(\s*(1\'.*)\),\s+\.CP\(\s*($SIGNAL_NAME\s*)\),\s+\.(S|C)D\(($SIGNAL_NAME\s*)\)/){
		    $inp1 = $1;
		    $ports =    $2 . ", " . $3 . ", " . "DONTCARE";	
		}
	    }else{
		if($line !~ /\.Q\(/) {
		    if ($line =~ /\.D\(\s*(1\'.*)\),\s+\.CP\(\s*($SIGNAL_NAME\s*)\),\s+\.(S|C)D\(($SIGNAL_NAME\s*)\),\s+/){
			$part1 =   $2 . ", " . $3;
		    }
		    if ($line =~ /\.QN\(\s*($SIGNAL_NAME\s*)\).*\)/) {
			$ports = $part1 . ", " . $1 . ", " . "DONTCARE";
		    }
		}
		if($line !~ /\.QN\(/) {
		    if ($line =~ /\.D\(\s*(1\'.*)\),\s+\.CP\(\s*($SIGNAL_NAME\s*)\),\s+\.(S|C)D\(($SIGNAL_NAME\s*)\),\s+\.Q\(($SIGNAL_NAME\s*)\)\s*/){
			$ports =  $2 . ", " . $3 . ", " . $4 .  ", " . "DONTCARE";
		    }
		}
	    }
	}
    }
    return $ports;
}

#
#
#
sub processDCVerLatches{
    my $line = shift;

    my $i = 0;
    my $port_list;
    my $outline = undef;
    if($line =~ /\'/){
	if($line =~ /($INSTANCE_NAME)\s*($PORTLIST);/){
	    #print "1: $1, 2:$2\n";
	    #print "match = $&\n";
	    $port_list = $2;
	    #print "port_list = $port_list\n";
	    print "WARNING: Line contain hard-coded assignment: $line";
	    $outline = parseHARDCODE($port_list);
	}
    } else{
	if ($line =~ /$DFF1\s*($ExtractVars::INSTANCE_NAME)\s*($PORTLIST);/){	
	    $port_list = $2;
	    #print "print $1 -- $2 -- $3\n";
	    $outline = parseDFF1($i,$port_list);
	}else{
	    if ($line =~ /$DFF2\s*($INSTANCE_NAME)\s*($PORTLIST);/){	
		$port_list = $2;
		$outline = parseDFF2($i,$port_list);
		#print "outline is $outline\n";
		if($outline eq '') { print "Warning: The following instance has no outputs:\nWarning:$line";}
	    }else{
		if ($line =~ /$DFF4\s*($INSTANCE_NAME)\s*($PORTLIST);/){	
		    $port_list = $2;
		    $outline = parseDFF4($i,$port_list);
		    if($outline eq '') { print "Warning: The following instance has no outputs:\nWarning:$line";}
		}
	    }
	}
    }
    return $outline;
}
#
# Find what is the reset net
#
sub defineReset{
    my $self = shift;
    my $FHO;
    
    if(! -e ".work/reset_ID"){
	print "BACKSPACE: What is the reset net in the techmapped design? ";
	chomp($RESET = <STDIN>);
	if(!$RESET) { die "ERROR: You must enter the reset net...Exiting\n";}
	open($FHO,">.work/reset_ID");
	print $FHO "$RESET";
	close($FHO);
    }else{
	open($FHO,"<.work/reset_ID");
	chomp($RESET = readline($FHO));
	close($FHO);
    }
    return 1;
}
