#CVS: $Id: GenerateDimacs.pm,v 1.5 2009/05/12 01:03:42 depaulfm Exp $

# LICENSE AGREEMENT
#
#  FileName                   [GenerateDimacs.pm]
#
#  PackageName                [BackSpace]
#
#  Synopsis                   [Generate Clauses on Dimacs fromat; generation is ]
#                             [based on Tseitin transforms                      ]
#
#  Author                     [FLAVIO M DE PAULA]
#
#  Copyright                  [Copyright (C) 2009 University of British Columbia]
# 
#  This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

package GenerateDimacs;

use 5.008;
use strict;
use warnings;
use Getopt::Long;
use Pod::Usage;
use File::Basename;
use Switch;

use lib "$ENV{'BS_PERL_MODULES'}";

use MyFileHandlers;
use ExtractVars;
use Benchmark;




###
# Private data to this class 
###
use vars qw ($logic
             $combVarsHandler
             );

BEGIN {
    use Exporter   ();
    use vars       qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);
    
    $VERSION = '0.01';
    
    @ISA         = qw(Exporter);
    @EXPORT      = qw(&generateDimacs &addUnitClauses); #export functions e.g. qw(&func1 &func2 &func4);
    %EXPORT_TAGS = ( );                            # eg: TAG => [ qw!name1 name2! ],
    
    # your exported package globals go here,
    # as well as any optionally exported functions
    @EXPORT_OK   =  qw(&readHash); #qw($Var1 %Hashit &func3);
}
use vars      @EXPORT_OK;

######################################
# Class constructor
#
######################################
sub new {
  my $package = shift;
  return bless({}, $package);
}

# initialize private variables
$logic = 0;     
$combVarsHandler = new ExtractVars;


######################################## 
# SUBROUTINES
########################################
#
#
#
sub generateDimacs{
    my $self = shift;
    my $arr_ptr = shift;
    my $hsh_ptr = shift;

    my $clauses_ref = tseitinTransforms($arr_ptr, $hsh_ptr);
    my @clauses     = addUnitClauses($arr_ptr, $hsh_ptr, $clauses_ref);

    my $out_file = ".work/clauses";
    my $FHO;
    open($FHO,">$out_file") || die "Cannot write file $out_file\n";
    print "... Generating clauses\n";
    for(my $i =0 ; $i <=$#clauses; $i = $i + 1){
	$clauses[$i] .= " 0\n";
	print $FHO "$clauses[$i]";
    }

    return @clauses;
}

#
#
#
sub readHash{
    my $self = shift;
    my $ptr = shift;
    my $hptr= shift;

    my @to_hash = @$ptr;
    my @splitter = ();
    for(my $i = 0; $i <= $#to_hash; $i = $i + 1){
	@splitter = split(' ', $to_hash[$i]);
	chomp $splitter[1]; # remove end of line
	$hptr->{$splitter[0]} = $splitter[1]; #used to be %$hptr yielding msgs: Using a hash as a reference is deprecated
    }
}
#
#
#
sub tseitinTransforms{
    my $arr_ptr = shift;
    my $hsh_ptr1 = shift;

    my @file_bare_bones = @$arr_ptr;
    my $i;
    my $line;
    my @tmp;
    my @clauses = ();
    my @result = ();
    for($i = 0; $i <= $#file_bare_bones; $i = $i + 1){
	if(($file_bare_bones[$i] =~ /\b$ExtractVars::AND2\b/) || 
	   ($file_bare_bones[$i] =~ /\b$ExtractVars::AND3\b/) ||
	   ($file_bare_bones[$i] =~ /\b$ExtractVars::AND4\b/))
	{
	    $line = $file_bare_bones[$i];
	   #print "Do I get here line[$i]: $line\n"; 
#	    $t0 = new Benchmark;
	    @tmp = $combVarsHandler->getCombVars($line);
#	    $t1 = new Benchmark;
#	    $td = timediff($t1,$t0);
#	    print "---> getCombVars " . timestr($td) . "\n";
	    @result =  processAND($hsh_ptr1, \@tmp);
	    push @clauses, @result;
if($logic){
	    print "line[$i]: $line\n";
	    foreach my $clause (@result) {
		print "$clause\n";
	    }
}
	    @tmp = ();
	}elsif(($file_bare_bones[$i] =~ /\b$ExtractVars::NAND2\b/) ||
	       ($file_bare_bones[$i] =~ /\b$ExtractVars::NAND3\b/) ||
	       ($file_bare_bones[$i] =~ /\b$ExtractVars::NAND4\b/) ||
	       ($file_bare_bones[$i] =~ /\b$ExtractVars::NAND5\b/)){
	    $line = $file_bare_bones[$i];
#	    $t0 = new Benchmark;
	    @tmp = $combVarsHandler->getCombVars($line);
#	    $t1 = new Benchmark;
#	    $td = timediff($t1,$t0);
#	    print "---> getCombVars " . timestr($td) . "\n";
	    @result = processNAND($hsh_ptr1, \@tmp);
	    push @clauses, @result;
if($logic){

	    print "line[$i]: $line\n";
	    foreach my $clause (@result) {
		print "$clause\n";
	    }
}
	    @tmp = ();
	}elsif(($file_bare_bones[$i] =~ /\b$ExtractVars::OR2\b/) ||
	       ($file_bare_bones[$i] =~ /\b$ExtractVars::OR3\b/) ||
	       ($file_bare_bones[$i] =~ /\b$ExtractVars::OR4\b/)){
	    $line = $file_bare_bones[$i];
#	    $t0 = new Benchmark;
	    @tmp = $combVarsHandler->getCombVars($line);
#	    $t1 = new Benchmark;
#	    $td = timediff($t1,$t0);
#	    print "---> getCombVars " . timestr($td) . "\n";
	    #foreach my $li (@tmp){ print "$line ==>$li\n";}
	    #foreach my $li (@tmp){ print "$li => $hsh_ptr1->{$li}\n";}

	    @result = processOR($hsh_ptr1, \@tmp);
	    push @clauses, @result;
if($logic){
	    print "line[$i]: $line\n";
	    foreach my $clause (@result) {
		print "$clause\n";
	    }
}
	    @tmp = ();
	}elsif(($file_bare_bones[$i] =~ /\b$ExtractVars::NOR2\b/) ||
	       ($file_bare_bones[$i] =~ /\b$ExtractVars::NOR3\b/) ||
	       ($file_bare_bones[$i] =~ /\b$ExtractVars::NOR4\b/) ||
	       ($file_bare_bones[$i] =~ /\b$ExtractVars::NOR5\b/)){
	    $line = $file_bare_bones[$i];
#	    $t0 = new Benchmark;
	    @tmp = $combVarsHandler->getCombVars($line);
#	    $t1 = new Benchmark;
#	    $td = timediff($t1,$t0);
#	    print "---> getCombVars " . timestr($td) . "\n";
	    @result = processNOR($hsh_ptr1, \@tmp);
	    push @clauses, @result;
if($logic){
	    print "line[$i]: $line\n";
	    foreach my $clause (@result) {
		print "$clause\n";
	    }
}
	    @tmp = ();
	}elsif($file_bare_bones[$i] =~ /\b$ExtractVars::NOT\b/){
	    $line = $file_bare_bones[$i];
#	    $t0 = new Benchmark;
	    @tmp = $combVarsHandler->getCombVars($line);
#	    $t1 = new Benchmark;
#	    $td = timediff($t1,$t0);
#	    print "---> getCombVars " . timestr($td) . "\n";
	    @result = processNOT($hsh_ptr1, \@tmp);
	    push @clauses, @result;
if($logic){
	    print "line[$i]: $line\n";
	    foreach my $clause (@result) {
		print "$clause\n";
	    }
}
	    @tmp = ();

	}
    }

    return \@clauses;
}

#
#
#
sub addUnitClauses{
#    my $self = shift;
    my $arr_ptr = shift;
    my $hsh_ptr1 = shift;
    my $arr_ptr2 = shift;

    my @file_bare_bones = @$arr_ptr;
    my $i;
    my $match;
    my @q = ();
    my @s = ();
    my @qn= ();
    my @sn= ();
    my $sign;
    my @clauses =@$arr_ptr2;

    #my $FH;
    #print "WARNING: Opening $out_file for appending, make sure this is only called once\n";
    #open($FH,">>$out_file") || die "Could not open $out_file to append binary clauses\n";
    for($i = 0; $i <= $#file_bare_bones; $i = $i + 1){
	if($file_bare_bones[$i] =~ /$ExtractVars::ALLDFF/){
	    if ($file_bare_bones[$i] =~ /\.D\(1\'b([0-1])\)/){
		$sign = $1;
		$sign =~ s/\s+//g;
		if($file_bare_bones[$i] =~ /\.Q\(([a-zA-Z0-9_]+\s*)\).*/){
		#if($file_bare_bones[$i] =~ /\.Q.*/){
		    $match = $1;
		    if(!defined($match)) { die "ERROR: $file_bare_bones[$i] does not match Q\n";}
		    $match =~ s/\s+//g;
		    push @q, $match;
		    push @s, $sign;
		}
		if ($file_bare_bones[$i] =~ /\.QN\(([a-zA-Z0-9_]+\s*)\).*\)/){
		#if($file_bare_bones[$i] =~ /\.QN.*/){
		    $match = $1;
		    if(!defined($match)) { die "ERROR: $file_bare_bones[$i] does not match QN\n";}
		    $match =~ s/\s+//g;
		    push @qn, $match;
		    push @sn, $sign;
		}
	    }
	}
    }

    for($i=0; $i <= $#q; $i= $i + 1){
	if($s[$i] == 0){
	    push @clauses, "-$hsh_ptr1->{$q[$i]}";
	}else{
	    push @clauses, "$hsh_ptr1->{$q[$i]}";
	}
	    
    }
    for($i=0; $i <= $#qn; $i= $i + 1){
	if($sn[$i] == 0){
	push @clauses, "$hsh_ptr1->{$qn[$i]}";
	}else{
	    push @clauses, "-$hsh_ptr1->{$qn[$i]}";
	}
    }
 #   for($i=0;$i <= $#clauses; $i = $i + 1){
#	print $FH $clauses[$i] . "\n";
#    }
#	print $FH "-$hsh_ptr1->{'GND'} 0\n";
#	print $FH "$hsh_ptr1->{'VCC'} 0\n";
    push @clauses, "-$hsh_ptr1->{'GND'}";
    push @clauses, "$hsh_ptr1->{'VCC'}";

   # close ($FH);
    return @clauses;
}

# Tseitin tranform:
# AND(o, i1, i2, ..., in) is equivalent to
# (i1 -o) 
# (i2 -o) ... 
# (in -o) 
# (-i1 -i2 ... -in o)
sub processAND{
    my $hsh_ptr1 = shift;
    my $array_ptr = shift;

    my @port_list = @$array_ptr;

    my @clauses = ();
    my $i;
    my $single;
    my $multiple = '';

    for($i=0;$i<$#port_list;$i=$i+1){
	#print "1:$hsh_ptr1->{$port_list[$i]}\n";
	$single =  "$hsh_ptr1->{$port_list[$i]} " . "-" . "$hsh_ptr1->{$port_list[$#port_list]}";
	$multiple .=  "-" . "$hsh_ptr1->{$port_list[$i]} ";
	push @clauses, $single;
    }
    $multiple .=  "$hsh_ptr1->{$port_list[$#port_list]}";

    push @clauses, $multiple;
    return @clauses;
}
# Tseitin tranform:
# OR(o, i1, i2, ..., in) is equivalent to
# (-i1 o) 
# (-i2 o) ... 
# (-in o) 
# (i1 i2 ... in -o)
sub processOR{
    my $hsh_ptr1 = shift;
    my $array_ptr = shift;

    my @port_list = @$array_ptr;

    my @clauses = ();
    my $i;
    my $single;
    my $multiple = '';

    for($i=0;$i<$#port_list;$i=$i+1){
	#print "1:$hsh_ptr1->{$port_list[$i]}\n";
	$single =   "-" . "$hsh_ptr1->{$port_list[$i]} " . "$hsh_ptr1->{$port_list[$#port_list]}";
	$multiple .=  "$hsh_ptr1->{$port_list[$i]} ";
	push @clauses, $single;
    }
    $multiple .=  "-" . "$hsh_ptr1->{$port_list[$#port_list]}";
    
    push @clauses, $multiple;
    return @clauses;
}
# Tseitin tranform: 
# NAND(o, i1, i2, ..., in) is equivalent to
# (i1 o) 
# (i2 o) ... 
# (in o) 
# (-i1 -i2 ... -in -o)
sub processNAND{
    my $hsh_ptr1 = shift;
    my $array_ptr = shift;

    my @port_list = @$array_ptr;

    my @clauses = ();
    my $i;
    my $single;
    my $multiple = '';

    for($i=0;$i<$#port_list;$i=$i+1){
	#print "1:$hsh_ptr1->{$port_list[$i]}\n";
	$single =  "$hsh_ptr1->{$port_list[$i]} " .  "$hsh_ptr1->{$port_list[$#port_list]}";
	$multiple .=  "-" . "$hsh_ptr1->{$port_list[$i]} ";
	push @clauses, $single;
    }
    $multiple .=  "-" . "$hsh_ptr1->{$port_list[$#port_list]}";    

    push @clauses, $multiple;
    return @clauses;
}
# Tseitin tranform: 
# NOR(o, i1, i2, ..., in) is equivalent to
# (-i1 -o) 
# (-i2 -o) ... 
# (-in -o) 
# (i1 i2 ... in o)
sub processNOR{
    my $hsh_ptr1 = shift;
    my $array_ptr = shift;

    my @port_list = @$array_ptr;

    my @clauses = ();
    my $i;
    my $single;
    my $multiple = '';

    for($i=0;$i<$#port_list;$i=$i+1){
	#print "1:$hsh_ptr1->{$port_list[$i]}\n";
	$single =  "-" . "$hsh_ptr1->{$port_list[$i]} " .  "-" . "$hsh_ptr1->{$port_list[$#port_list]}";
	#print "==$hsh_ptr1->{$port_list[$i]}==$multiple==\n";
	$multiple .=  "$hsh_ptr1->{$port_list[$i]} ";
	#print "==$hsh_ptr1->{$port_list[$i]}==$multiple==\n";
	push @clauses, $single;
    }
    $multiple .=  "$hsh_ptr1->{$port_list[$#port_list]}";
    
    push @clauses, $multiple;
    return @clauses;
}
# Tseitin tranform:
# NOT(o, i1) is equivalent to
# (-i1 -o) 
# (i1 o)
sub processNOT{
    my $hsh_ptr1 = shift;
    my $array_ptr = shift;

    my @port_list = @$array_ptr;

    my @clauses = ();
    my $i;

    my $first =  "-" . "$hsh_ptr1->{$port_list[1]} " . "-" . "$hsh_ptr1->{$port_list[0]}";
    push @clauses, $first;
    my $second=  "$hsh_ptr1->{$port_list[1]} " . " $hsh_ptr1->{$port_list[0]}";
    push @clauses, $second;
    return @clauses;
}
