#CVS: $Id: GenerateInstance.pm,v 1.8 2009/05/12 01:03:42 depaulfm Exp $

# LICENSE AGREEMENT
#
#  FileName                   [GenerateInstance.pm]
#
#  PackageName                [BackSpace]
#
#  Synopsis                   [Generate problem instance for the SAT solver     ]
#
#  Author                     [FLAVIO M DE PAULA]
#
#  Copyright                  [Copyright (C) 2009 University of British Columbia]
# 
#  This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

package GenerateInstance;

use 5.008;
use strict;
use warnings;
use Getopt::Long;
use Pod::Usage;
use File::Basename;
use Switch;

use lib "$ENV{'BS_PERL_MODULES'}";

use MyFileHandlers;
use ExtractVars;


###
# Private data to this class 
###
use vars qw ($setOfStatesF
             $setOfConstrF
             $target
             $constrain
             $total
             $fraction
             $careSetF
             );

BEGIN {
    use Exporter   ();
    use vars       qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);
    
    $VERSION = '0.01';
    
    @ISA         = qw(Exporter);
    @EXPORT      = qw(&instantiate); #export functions e.g. qw(&func1 &func2 &func4);
    %EXPORT_TAGS = ( );                            # eg: TAG => [ qw!name1 name2! ],
    
    # your exported package globals go here,
    # as well as any optionally exported functions
    @EXPORT_OK   =  qw(); #qw($Var1 %Hashit &func3);
}
use vars      @EXPORT_OK;

######################################
# Class constructor
#
######################################
sub new {
  my $package = shift;
  return bless({}, $package);
}

# initialize private variables
$setOfStatesF = undef;     # 
$setOfConstrF = undef;
$target       = 1;
$constrain    = 0;
$total        = 1;
$fraction     = 16;
$careSetF     = undef;



######################################## 
# SUBROUTINES
########################################
#
# Main routine being exported
#
sub instantiate{
    my $self = shift;
    my $clauses_ptr      = shift;
    my $symTable_ptr     = shift;
    my $cleaned_ptr      = shift;
    my $constraint_fname = shift;
    my $target_fname     = shift;
    my $careSet_fname    = shift;

    # Set private data
    $setOfConstrF = $constraint_fname;
    $setOfStatesF = $target_fname;
    $careSetF     = $careSet_fname;

    my $out_file;
    for(my $Index=0; $Index < $total; $Index = $Index + 1){
	$out_file = "instance" . $Index . ".cnf";
	outputDIMACS($clauses_ptr, $symTable_ptr, $cleaned_ptr, $out_file, $Index);
    }

    my $careH = new MyFileHandlers;
    $careH->MFHreadfile($careSetF);
    my @care_set = $careH->MFHfile;
    my $indexes_ptr = filterLatches($cleaned_ptr,\@care_set);
    my @indexes = @$indexes_ptr;
    my @latches = getLatchesOutCareSet($cleaned_ptr,\@indexes);
    generateCHeader(\@latches, $symTable_ptr);
    my $size = $#latches + 1;
    my @constr_indexes =  filterLatchesPos($cleaned_ptr,\@care_set);
    for(my $Index=0; $Index < $total; $Index = $Index + 1){
 	generateConstraintCareSet($symTable_ptr, $cleaned_ptr, \@indexes,\@constr_indexes, $size, $Index);
    }

}

#
#
#
sub outputDIMACS{
    my $arr_ptr  = shift;
    my $hsh_ptr1 = shift;
    my $arr_ptr2 = shift;
    my $out_file = shift;
    my $index    = shift;

    my @clauses = @$arr_ptr;
    my %vars = %$hsh_ptr1;
    my @file_bare_bones = @$arr_ptr2;

    my $keyed;
    my $FH;
    print "Opening $out_file\n";
    open($FH,">$out_file") || die "Could not open $out_file for writing\n";
    
    my $latch = 0;
    my $offset;
    my $divd;
    my $nconstr;
    my $cnt_clauses = $#clauses +  (keys ( %vars ));
    print $FH "c\n";
    print $FH "p cnf " . keys ( %vars ) . " $cnt_clauses\n";
    foreach my $clause (@clauses) {
	print $FH "$clause"; ### assume $clause format <lit [{lit}] 0\n> 
    }
     my $q;
    my $qn;
    #
    # Some latches have Q and QN. This pair of signals have to follow the
    # Tseitin's  NOT transformation
	for(my $i = 0; $i <= $#file_bare_bones; $i = $i + 1){
	    if($file_bare_bones[$i] !~ /\'/){
		if($file_bare_bones[$i] =~ /$ExtractVars::ALLDFF/){
		    if(($file_bare_bones[$i] =~ /\.Q\(\s*($ExtractVars::SIGNAL_NAME)\s*\)/) &&
		       ($file_bare_bones[$i] =~ /\.QN\(\s*($ExtractVars::SIGNAL_NAME)\s*\)/)){ 
			if($file_bare_bones[$i] =~ /\.Q\(\s*($ExtractVars::SIGNAL_NAME)\s*\)/){ 
			    if(defined($1)) {
				$q = $1;
				$q=~ s/\s+//g;
			    }
			}
			if($file_bare_bones[$i] =~ /\.QN\(\s*($ExtractVars::SIGNAL_NAME)\s*\)/){ 
			    if(defined($1)) {
				$qn = $1;
				$qn =~ s/\s+//g;
			    }
			} 
			print $FH "$vars{$q} $vars{$qn} 0\n";
			print $FH "-" . $vars{$q} . " -" . $vars{$qn} . " 0\n";
		    }
		}
	    }
	}

    ##
    ## Only interested in pre-img so I will parse the last item in the portlist, which 
    ## has the latch input 'D'. In SAT, we want to constrain 'D' since it is the output of
    ## the combinational circuit prior to the latch
    ##
    $nconstr = 0;
    if($constrain){
	my @constr = readState('constraints',$index);
 	for(my $i = 0; $i <= $#file_bare_bones; $i = $i + 1){
	    if(($file_bare_bones[$i] =~ /\.Q\(\s*($ExtractVars::SIGNAL_NAME)\s*\)/) ||
	       ($file_bare_bones[$i] =~ /\.QN\(\s*($ExtractVars::SIGNAL_NAME)\s*\)/)){ 
		$nconstr = $nconstr + 1;
	    }
	}
	for(my $i = 0; $i <= $#file_bare_bones; $i = $i + 1){
	    if($file_bare_bones[$i] !~ /\'/){
		if($file_bare_bones[$i] =~ /$ExtractVars::ALLDFF/){
		    if(!defined($constr[$latch])){die "ERROR: constr[$latch] not defined\n";}
		    if($file_bare_bones[$i] =~ /\.Q\(\s*($ExtractVars::SIGNAL_NAME)\s*\)/){ 
			if(defined($1)) {
			    $keyed = $1;
			    $keyed =~ s/\s+//g;
			    if($constr[$latch] eq 0){
			    print $FH "-" . $vars{$keyed} . " 0\n";
			}
			    else{
				print $FH $vars{$keyed} .  " 0\n";
			    }
			}
		    }
		    if($file_bare_bones[$i] =~ /\.QN\(\s*($ExtractVars::SIGNAL_NAME)\s*\)/){ 
			if(defined($1)) {
			    $keyed = $1;
			    $keyed =~ s/\s+//g;
			}
			#Because it's QN, the actual output is  negated
			if($constr[$latch] eq 1){
			    print $FH "-" . $vars{$keyed} . " 0\n";
			}
			else{
			    print $FH $vars{$keyed} . " 0\n";
			}
		    }
		    $latch = $latch + 1;
		    $divd = $nconstr / $fraction;
		    if($latch >= $divd){
			last;
		    }
		}
	    }
	}
    }
    if($target){
	my @state  = readState('states',$index);
	$latch = 0;
	#
	# Add the State
	#
	for(my $i = 0; $i <= $#file_bare_bones; $i = $i + 1){
	    if($file_bare_bones[$i] !~ /\'/){
		if($file_bare_bones[$i] =~ /$ExtractVars::ALLDFF/){
		    if($file_bare_bones[$i] =~ /\.D\(\s*($ExtractVars::SIGNAL_NAME)\s*\)/){ 
			if(defined($1)) {
			    $keyed = $1;
			    $keyed =~ s/\s+//g;
			}else{ die " something wrong when parsing $ExtractVars::ALLDFF\n";}
			if($state[$latch] == 0){
			    print $FH "-" . $vars{$keyed} . " 0\n";
			}
			else{
			    print $FH $vars{$keyed} . " 0\n";
			}
		    }
		    $latch = $latch + 1;
		}
	    }
	}	
    }#target
	    
    #
    # Add reset 
    #
    my $getResetID = new ExtractVars;
    $getResetID->defineReset();
#	print $FH "-" . $vars{'n555'}	. " 0\n";     
    print $FH "-" . $vars{$ExtractVars::RESET}	. " 0\n";  
    close($FH);  
    
}
#
#
#
sub readState{
    my $type = shift;
    my $Index =shift;
    my $FH;
    my @list = (0);
    my $line = undef;
    my $previousLine = undef;
    my @temp1 = ();
    my @localIndx = ();
    my @state = ();
    if($type =~ /states/){ 
	if(defined($setOfStatesF)) {print "Opening file $setOfStatesF\n";} else {die "File setOfstates was not defined\n";}
	open($FH,"<$setOfStatesF") || die "Error: can not open file $setOfStatesF";
    }else{
	if(defined($setOfConstrF)) {print "Opening file $setOfConstrF\n";} else {die "File setOfConstr was not defined\n";}
	open($FH,"<$setOfConstrF") || die "Error: can not open file $setOfConstrF";
    }

    my $stateIndx = $Index;
    while(defined($line = readline($FH))) {
	my $pat = 'state[\[0-9]+\]\s\=';
	if(($line =~ $pat)) { # && ($line !~ /state\[0\]\s\=/)) {
	    $previousLine = $line;
	}elsif(defined($previousLine)) {
	    @temp1 = split('\[',$previousLine);
	    @localIndx = split('\]',$temp1[1]);
	    if($localIndx[0] == $stateIndx){
		$previousLine = undef;
		@state = split(',', $line);
		last;
	    }
	}
    }
    close ($FH);
    return @state;

}


####
# REQUIREMENT ::     
#    $careH = new MyFileHandlers;
#    $careH->MFHreadfile($careSetF);
#    my @care_set = $careH->MFHfile;
# call: filterLatches(\@file_bare_bones,\@care_set)
##
sub filterLatches{
    my $arr_ptr1 = shift; # All latches
    my $arr_ptr2 = shift; # careSet

    my @file_bare_bones = @$arr_ptr1;
    my @care_set = @$arr_ptr2;
    my $j;
    my @each_latch;
    my @filtered_latches_indexes;
 
    my  $filterH = new MyFileHandlers;
    my $currdir = $filterH->MFHcurdir;
    my $workdir = $filterH->MFHconcatdir2($currdir,".work");
    my $filtered_file   = $filterH->MFHconcatfile($workdir,"state_filter");
 
    my $temp_each_latch;
    # If file doesn't exist, generate one, otherwise read the existing file
    if(! -e $filtered_file){
	for(my $i=0; $i <= $#file_bare_bones; $i = $i +1){
	    if($file_bare_bones[$i] =~ /$ExtractVars::ALLDFF/){
		if ($file_bare_bones[$i] !~ /1\'/){
		    $j = 0;
		  search_set: while($j <= $#care_set){
		      @each_latch = split(/\s+/,$care_set[$j]);
		      $each_latch[0] =~ s/\[/\\[/g;###### solves problem w/ regexp string w/ [] and [][]
		      $each_latch[0] =~ s/\]/\\]/g;
		      if($file_bare_bones[$i] =~ "$each_latch[0]"){
			  push @filtered_latches_indexes, $i . "\n";
			  last search_set;
		      }
		      $j = $j + 1; 
		  }
		}
	    }
	}
	$filterH->MFHwritefile($filtered_file,\@filtered_latches_indexes);
   } else{
	$filterH->MFHreadfile($filtered_file);
	@filtered_latches_indexes = $filterH->MFHfile;
    }
    return \@filtered_latches_indexes;
}

#
#
#
sub filterLatchesPos{

    my $arr_ptr1 = shift; # All latches
    my $arr_ptr2 = shift; # careSet


    my @file_bare_bones = @$arr_ptr1;
    my @care_set = @$arr_ptr2;
    my $j;
    my @each_latch;
    my @filtered_latches_indexes;
 
    my $filterH = new MyFileHandlers;
    my $currdir = $filterH->MFHcurdir;
    my $workdir = $filterH->MFHconcatdir2($currdir,".work");
    my $filtered_file   = $filterH->MFHconcatfile($workdir,"constraint_indexes");


    # If file doesn't exist, generate one, otherwise read the existing file
    if(! -e $filtered_file){
	for(my $i=0; $i <= $#file_bare_bones; $i = $i +1){
	    if($file_bare_bones[$i] =~ /$ExtractVars::ALLDFF/){
		if ($file_bare_bones[$i] !~ /1\'/){
		    $j = 0;
		  search_set: while($j <= $#care_set){
		      @each_latch = split(' ',$care_set[$j]);
		      $each_latch[0] =~ s/\[/\\[/g;
		      $each_latch[0] =~ s/\]/\\]/g;
		      if($file_bare_bones[$i] =~ /$each_latch[0]/){
			  push @filtered_latches_indexes, $each_latch[1] . "\n";
			  last search_set;
		      }
		      $j = $j + 1; 
		  }
		}
	    }
	}
	$filterH->MFHwritefile($filtered_file,\@filtered_latches_indexes);
   } else{
	$filterH->MFHreadfile($filtered_file);
	@filtered_latches_indexes = $filterH->MFHfile;
    }
    return @filtered_latches_indexes;

}
#
#
#
sub generateCHeader{
   my $arr_ptr = shift;
   my $hsh_ptr = shift;
   my @latches = @$arr_ptr;
   my $name = "cnfLatches.h";
   
   my  $cnfLatchesH = new MyFileHandlers;
   my $currdir  = $cnfLatchesH->MFHcurdir;
   my $out_file = $cnfLatchesH->MFHconcatfile($currdir, $name);

   my $FH;
   my $i;
    print "Opening $out_file\n";
   open($FH,">$out_file") || die "Could not open $out_file";

   print $FH "/*###################################################
File automaticaly generated from $0
###################################################*/\n";
   my $size = $#latches + 1;
   print $FH "unsigned int nlatches = " . $size  . ";\n";
   
   print $FH "unsigned int latches[$size] = {\n";
   for($i=0; $i < $#latches; $i = $i +1) {
       print $FH $hsh_ptr->{$latches[$i]};
       print $FH ",";
   }
   print $FH $hsh_ptr->{$latches[$i]};
   print $FH "\n};\n"
}
#
#
#
sub getLatchesOutCareSet{
    my $arr_ptr1 = shift;
    my $arr_ptr2 = shift;


    my @file_bare_bones = @$arr_ptr1;
    my @indexes         = @$arr_ptr2;


    my $latch = undef;
    my $j;
    my @each_latch;
    my @latch_array;
 

		#$t0 = new Benchmark; # 30s

    for(my $i=0; $i <= $#indexes; $i = $i +1){
	#Considers either Q or QN
	if($file_bare_bones[$indexes[$i]] =~ /\.Q\(\s*($ExtractVars::SIGNAL_NAME\s*)\)/){
	    $latch = $1;
	    if(defined($latch)){
		$latch =~ s/\s+//g;
		#print "$latch\n";
		push @latch_array, $latch;
	    }
	}else{
	    if($file_bare_bones[$indexes[$i]] =~ /\.QN\(\s*($ExtractVars::SIGNAL_NAME\s*)\)/){
		$latch = $1;
		if(defined($latch)){
		    $latch =~ s/\s+//g;
		    #print "$latch\n";
		    push @latch_array, $latch;
		}
	    }
	}
    }

		#$t1 = new Benchmark;
		#$td = timediff($t1,$t0);
		#print "---> getLatchesOutCareSet : elapsed time is " . timestr($td) . "\n";
    return @latch_array;
}
#
#
#
sub generateConstraintCareSet{
    my $hsh_ptr1 = shift;
    my $arr_ptr2 = shift;  # cleaned.v 
    my $arr_ptr3 = shift;  # indexes of of latches in cleaned.v
    my $arr_ptr4 = shift;  # indexes of of latches in constraint.h
    my $size     = shift;      # number of latches
    my $Index    = shift;
    
    my %vars            = %$hsh_ptr1;
    my @file_bare_bones = @$arr_ptr2;
    my @indexes         = @$arr_ptr3;
    my @constr_indexes  = @$arr_ptr4;

    my $keyed;
    my $FH;
    my $name = "constraints" . $Index . ".h";
    my  $constraintsH = new MyFileHandlers;
    my $currdir  = $constraintsH->MFHcurdir;
    my $out_file =  $constraintsH->MFHconcatfile($currdir,$name);
    print "Opening $out_file\n";
    open($FH,">$out_file") || die "Could not open $out_file for writing\n";
    
    my $latch = 0;

    my @constr = readState('constraints',$Index);

    my @list = ();


    print $FH "/*###################################################
File automaticaly generated from $0
###################################################*/\n";
	#	$t0 = new Benchmark; # 26
    for(my $i = 0; $i <= $#indexes; $i = $i + 1){
	if(!defined($constr[$constr_indexes[$i]])){die "ERROR: constr[$latch] not defined\n";}
	# Considers either Q or QN
	if($file_bare_bones[$indexes[$i]] =~ /\.Q\(\s*($ExtractVars::SIGNAL_NAME)\s*\)/){ 
	    if(defined($1)) {
		$keyed = $1;
		$keyed =~ s/\s+//g;
		push @list, $constr[$constr_indexes[$i]];
	    }
	}else{
	    if($file_bare_bones[$indexes[$i]] =~ /\.QN\(\s*($ExtractVars::SIGNAL_NAME)\s*\)/){ 
		if(defined($1)) {
		    $keyed = $1;
		    $keyed =~ s/\s+//g;
		    my $val =  ($constr[$constr_indexes[$i]]) ? 0 : 1;
		    push @list, $val;
		}
	    }
	}
    }

		#$t1 = new Benchmark;
		#$td = timediff($t1,$t0);
		#print "---> generateConstraintCareSet : elapsed time is " . timestr($td) . "\n";

    if($size ne ($#list+1)) {print  "WARNING: Number of Q/QN passed to generateCostraint differ from count: size=$size, list=$#list + 1 (Chosen the latter, computed locally)\n";}
    my $nsize= $#list + 1;

    print $FH "unsigned int misat_sign[$nsize] = {\n";
    my $i;
    for($i=0;$i<$#list;$i=$i+1){
	print $FH "$list[$i],";
    }
    print $FH "$list[$i]};\n";
    @list = ();
}


1;
