#CVS: $Id: GenerateSymTable.pm,v 1.4 2009/05/12 01:03:42 depaulfm Exp $

# LICENSE AGREEMENT
#
#  FileName                   [GenerateSymTable.pm]
#
#  PackageName                [BackSpace]
#
#  Synopsis                   [Uses ExtractVars to generate symbol table]
#
#  Author                     [FLAVIO M DE PAULA]
#
#  Copyright                  [Copyright (C) 2009 University of British Columbia]
# 
#  This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

package GenerateSymTable;

use 5.008;
use strict;
use warnings;
use Getopt::Long;
use Pod::Usage;
use File::Basename;
use Switch;

use lib "$ENV{'BS_PERL_MODULES'}";

use MyFileHandlers;
use ExtractVars;


###
# Private data to this class 
###
use vars qw ($combVarsHandler
             $latchVarsHandler
             $global_cnt
             );

BEGIN {
    use Exporter   ();
    use vars       qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);
    
    $VERSION = '0.01';
    
    @ISA         = qw(Exporter);
    @EXPORT      = qw(&getVariables &readHash); #export functions e.g. qw(&func1 &func2 &func4);
    %EXPORT_TAGS = ( );                            # eg: TAG => [ qw!name1 name2! ],
    
    # your exported package globals go here,
    # as well as any optionally exported functions
    @EXPORT_OK   =  qw(); #qw($Var1 %Hashit &func3);
}
use vars      @EXPORT_OK;

######################################
# Class constructor
#
######################################
sub new {
  my $package = shift;
  return bless({}, $package);
}

# initialize private variables
$combVarsHandler  = new ExtractVars;
$latchVarsHandler = new ExtractVars;
$global_cnt = 1;



######################################## 
# SUBROUTINES
########################################
#
#
#
sub getVariables{
    my $self  = shift;
    my $arr_ptr  = shift;

    my @file_bare_bones = @$arr_ptr;
    my %latch_vars; #old $hsh_ptr1
    my %comb_vars; #old $hsh_ptr2

    my $i;
    my $line;
    my @tmp = ();
    my $all_cells  = $ExtractVars::CELL;
    my $only_dffs  = $ExtractVars::ALLDFF;

    my $reset_defined = 0;
    ## Try to automatically figure out the reset net
    if($reset_defined == 0){
        $reset_defined = $latchVarsHandler->defineReset();
    }
    for($i = 0; $i <= $#file_bare_bones; $i = $i + 1){
	if($file_bare_bones[$i] =~ /$all_cells/){
	    if($file_bare_bones[$i] =~ /$only_dffs/){
		$line = $file_bare_bones[$i];

		@tmp =  $latchVarsHandler->getLatchVars($line);
		foreach my $entry (@tmp){
		    if((!defined($latch_vars{$entry})) && (!defined($comb_vars{$entry}))){
			$latch_vars{$entry} = $global_cnt;
			$global_cnt = $global_cnt + 1;
		    }
		}
#		foreach my $entry (@tmp){
#		    print "entry $entry\n"
#		}
		@tmp = ();
	    }else{
		$line = $file_bare_bones[$i];
		@tmp = $combVarsHandler->getCombVars($line);
		foreach my $entry (@tmp){
		    if((!defined($latch_vars{$entry})) && (!defined($comb_vars{$entry}))){
			$comb_vars{$entry} = $global_cnt;
			$global_cnt = $global_cnt + 1;
		    }
		}
#		foreach my $entry (@tmp){
#		    print "entry $entry\n"
#		}
		@tmp = ();
	    }
	}#//CELL
    }#//for

#
# Save the symbol table into a file and return it 
#
    my %vars;
    my $keyed;
    my @sorted;
    my $val;
    my %symTable = ();
    foreach $keyed (keys %latch_vars){
	$vars{$keyed} = $latch_vars{$keyed};
    }
    foreach $keyed (keys %comb_vars){
	$vars{$keyed} = $comb_vars{$keyed};
    }
    
    @sorted = sort { $vars{$a} <=> $vars{$b} } keys %vars;
    print "... Generating Symbol Table\n";
    my $FHO;    my $out_file =".work/symTable";
    open($FHO,">$out_file") || die "Cannot open $out_file";
    foreach $val (@sorted){
	print $FHO "$val $vars{$val}\n";
	$symTable{$val} = $vars{$val}; ## does not guarantee the same order as 
                                       ## printed in hashTable
    }
    close($FHO);
    return \%symTable;
}
#
#
#
sub readHash{
    my $self = shift;
    my $ptr = shift;
    my $hptr= shift;

    my @to_hash = @$ptr;
    my @splitter = ();

    for(my $i = 0; $i <= $#to_hash; $i = $i + 1){
	@splitter = split(' ', $to_hash[$i]);
	chomp $splitter[1]; # remove end of line
	%$hptr->{$splitter[0]} = $splitter[1];
    }

}
