#CVS: $Id: Sim2Sat.pm,v 1.2 2009/03/31 08:13:11 depaulfm Exp $

# LICENSE AGREEMENT
#
#  FileName                   [Sim2Sat.pm]
#
#  PackageName                [BackSpace]
#
#  Synopsis                   [Interface between simulation 'crash state dump   ]
#                             [and Dimacs' clauses generation to the SAT solver ]
#
#  Author                     [FLAVIO M DE PAULA]
#
#  Copyright                  [Copyright (C) 2009 University of British Columbia]
# 
#  This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

package Sim2Sat;

use 5.008;
use strict;
use warnings;
use Getopt::Long;
use Pod::Usage;
use File::Basename;
use Switch;

use lib "$ENV{'BS_PERL_MODULES'}";

use MyFileHandlers;
use ExtractVars;


###
# Private data to this class 
###
use vars qw ($constr
             $fileHnd
             );

BEGIN {
    use Exporter   ();
    use vars       qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);
    
    $VERSION = '0.01';
    
    @ISA         = qw(Exporter);
    @EXPORT      = qw(&generateFiles ); #export functions e.g. qw(&func1 &func2 &func4);
    %EXPORT_TAGS = ( );                            # eg: TAG => [ qw!name1 name2! ],
    
    # your exported package globals go here,
    # as well as any optionally exported functions
    @EXPORT_OK   =  qw( ); #qw($Var1 %Hashit &func3);
}
use vars      @EXPORT_OK;

######################################
# Class constructor
#
######################################
sub new {
  my $package = shift;
  return bless({}, $package);
}

# initialize private variables
$constr = 1;
$fileHnd = new MyFileHandlers;

#
#Main program
#
sub generateFiles{
    my $self = shift;
    my $vcsdir = shift;
    my $oudir     = shift;
    my $stateslog = shift;
    my $offlist = shift;
    my $verilog_ptr = shift;


#    print "$vcsdir $oudir $stateslog $offlist $verilog_ptr\n";
#    exit();

    my @list = ();
    my $countFFs = countDutFFs($verilog_ptr);
#    print "CountFFs = $countFFs\n";
    @list = writeTargetList($vcsdir, $oudir, $stateslog);
    
    genStates($vcsdir, $oudir, \@list, $countFFs);
    
    genCareSet($vcsdir, $oudir, $verilog_ptr, $offlist); ### modify so that infile is @cleaned also modify this fun guts

}
#########################################################
# SUBROUTINES
#########################################################
#
# Parse results and create a target.list file for each benchmark
#
sub writeTargetList{
    my ($vcsdir, $oudir, $patt) = @_;

    my @list = ();
    my @file_bare_bones = ();
    my $match;
    my @date = localtime(time);
    my $FH;

# Create absolute path and check that it exists
    my $currdir = $fileHnd->MFHcurdir;
    my $srcdir = $fileHnd->MFHconcatdir($currdir,$vcsdir);

# Name of the benchmark
    my @temp = split('\/',$vcsdir);
    my $bench = $temp[1];

# Create absolute path and check that it exists
    my $destdir = $fileHnd->MFHconcatdir2($currdir,$oudir);

# Create one directory for each benchmark
    my $eachdir = $fileHnd->MFHconcatdir2($destdir,$bench);
# Create file 
    my $targetlist = $fileHnd->MFHconcatfile($eachdir, "target.list");

#
# Create a list of files to be parsed and modified
#
    $fileHnd->MFHfind($srcdir,$patt);
    my @files = $fileHnd->MFHlistoffiles;

    my @pair_list= ();
    my $match_T; 
    my $match_C;
    my $matched_T = 0;
    my $matched_C = 0;
    my @list_C = ();
    my @list_T = ();
# For each file parse TARGET.* and
# push it into an array if it is unique 
    if($constr){
	#if CONSTRAINT and TARGET pairs, check for their uniqueness
	foreach my $file (@files){
	    print "Processing file $file\n";
	    # Get the file itself into an array
	    $fileHnd->MFHreadfile($file);
	    @file_bare_bones = $fileHnd->MFHfile;
	    foreach my $line (@file_bare_bones){
		if($line =~ /CONSTRAINT.*/) {
		    $match_C = $&;
		    chomp $match_C;
		    #print "Do I get Here : $match_C\n";
		}
		if(($line =~ /TARGET.*/)) {
		    $match_T = $&;
		    chomp $match_T;
		    #print "Do I get Here : $match_T\n";
		}
	    }
	}
	push @list, getSubstrings($match_C);
	push @list, getSubstrings($match_T);

	#print "C = $match_C\n";
	#print "T = $match_T\n";
	
#	push @list, $match_C;
#	push @list, $match_T;
    }else{
	die  "ERROR: Deleted branch when --constr is not passed";
    }

# Prepare to output target.list 
    my $header1 = "##########
#  This is an automatically generated target list based on VCS simulations in $vcsdir.
#
#  File created w/ the following command-line:\n";
    
   my $header2 ="\n#
#  File created at " . $date[4] ."/". $date[3] ."/". ($date[5] + 1900). "
###########\n";

   print "Creating file $targetlist \n";
   open($FH,">$targetlist") || die "Error: can not open file $targetlist";

# Output target.list    
    print $FH $header1;
#    for(my $i = 0; $i <=$#ARGV; $i = $i+1){
#	print $FH "$ARGV[$i]";
#    } 
    print $FH $header2;
#print "Found $#list distinct TARGETS\n";
    foreach my $line (@list){
	print $FH "$line\n";
	#if($line !~ /\)/){
	#    print $FH "$line";
	#}else{
	#     print $FH "$line\n";
	# }
    }
    close ($FH);

    return @list;
}

#
# Augment benchmark w/ RESET and TARGET states
#
sub genStates{
    my ($vcsdir, $oudir, $ptrlist, $flipflop) = @_;
    
# Name of the benchmark
    my @temp = split('\/',$vcsdir);
    my $bench = $temp[1];

# Create absolute path and check that it exists
    my $currdir = $fileHnd->MFHcurdir;
    my $destdir = $fileHnd->MFHconcatdir2($currdir,$oudir);

# Create one directory for each benchmark
    my $eachdir = $fileHnd->MFHconcatdir2($destdir,$bench); 
    my @list = @$ptrlist;
    my $iter = 0;
    
# Get the file itself into an array
    my @file_bare_bones = (); #

# Loop through each file and augment it 
    my $inst1 = 0; my $inst2 = 0;
    my $nstates = ($#list + 1) / 2;
    foreach my $line (@list){
	if($line =~ /TARGET/){
	    processTarget($nstates, $eachdir, "$bench.bench", $iter, $line, \@file_bare_bones, $flipflop);
	    $inst1 = 1;
	}elsif($line =~ /CONSTRAINT/){
	    processConstraint($nstates, $eachdir, "$bench.bench", $iter, $line, \@file_bare_bones, $flipflop);
	    $inst2 = 1;
	}
	if($inst1 == 1 && $inst2 == 1){
	    $iter = $iter + 1;
	    $inst1 = 0;$inst2 = 0;
	}
    }
}
#
# Subroutine responsible for the actual augmentation
#
sub processTarget{
    my ($nstates,$destdir, $file, $nth, $line, $bare_bones, $flipflop) = @_;
 
    my $i;
    my $FH;
#    my $flipflop = 0;

# Dereferencing arrays
    my @file_bare_bones = @$bare_bones;

    my $file_name = "setOfStates.h";
    my $abs_name = $fileHnd->MFHconcatfile($destdir,$file_name);

# Split $line to get only the array
    chop $line;
    my @array = split('\(',$line);
    my @arrayElems = split(',',$array[1]);

    #Creating .h header file 
    print "Creating file $abs_name \n";
    open($FH,">$abs_name") || die "Error: can not open file $abs_name";
  
#    print "nth=$nth\n";
    if($nth < 1){ # allows for both setOfConstraints and setOfStates 
	print $FH "
/*##########
#  This file is based on the original file $file.
#  It includes the RESET state (assumed to be all zeros) 
# and a pseudo-random TARGET state.
###########\n*/
";
	# Declaration of variable state
	print $FH "unsigned int state[$nstates][$flipflop] = {\n";
    }

# Output States
    print $FH "\n//state[$nth] =\n";
    for(my $i=0; $i <=$#arrayElems; $i=$i+1) {
	if($nth < ($nstates-1)){
	    print $FH "$arrayElems[$i],";
	}else{
	    if($i < $#arrayElems){
		print $FH "$arrayElems[$i],";
	    }else{
		print $FH "$arrayElems[$i]\n};\n";
		close ($FH);
	    }
	}
    }
}

#
# Subroutine responsible for the actual augmentation
#
sub processConstraint{
    my ($nstates,$destdir, $file, $nth, $line, $bare_bones,$flipflop) = @_;
 
    my $i;
    my $FH;
#    my $flipflop = 0;

# Dereferencing arrays
    my @file_bare_bones = @$bare_bones;

    my $file_name = "setOfConstraints.h";
    my $abs_name = $fileHnd->MFHconcatfile($destdir,$file_name);

# Split $line to get only the array
    chop $line;
    my @array = split('\(',$line);
    my @arrayElems = split(',',$array[1]);

    #Creating .h header file 
    print "Creating file $abs_name \n";
    open($FH,">$abs_name") || die "Error: can not open file $abs_name";
  
#    print "nth=$nth\n";
    if($nth < 1){ # allows for both setOfConstraints and setOfStates 
	print $FH "
/*##########
#  This file is based on the original file $file.
#  It includes the RESET state (assumed to be all zeros) 
# and a pseudo-random TARGET state.
###########\n*/
";
	# Declaration of variable state
	print $FH "unsigned int state[$nstates][$flipflop] = {\n";
    }

# Output States
    print $FH "\n//state[$nth] =\n";
    for(my $i=0; $i <=$#arrayElems; $i=$i+1) {
	if($nth < ($nstates-1)){
	    print $FH "$arrayElems[$i],";
	}else{
	    if($i < $#arrayElems){
		print $FH "$arrayElems[$i],";
	    }else{
		print $FH "$arrayElems[$i]\n};\n";
		close ($FH);
	    }
	}
    }
}



#
# Check if $match if unique in the @list 
#
sub checkUnique{
    my ($match,$ptrlist) = @_;
 
    my $i;
    my $FH;
    my $entry ='';
    my $result = 1;
    my @list = @$ptrlist;

#    print "size of list $#list == match passed is $match-\n";
    if($#list >=0){
	foreach $entry (@list){
	    if($match eq "$entry"){
#		print "compare $entry-$match-\n";
		$result = 0;
		last;
	    }
	}
    }
    return $result;
}



#
# Assumption: order of latches in the state follows order in which
#             latches are defined
#
sub genCareSet{
    my ($vcsdir, $oudir, $verilog_ptr, $offlist) = @_;
    
# Name of the benchmark
    my @temp = split('\/',$vcsdir);
    my $bench = $temp[1];

# Create one directory for each benchmark
    my $destdir = $fileHnd->MFHconcatdir2($oudir,$bench);

#    print "Reading file $inFile\n";
    my @file_bare_bones = @$verilog_ptr;

    my @off_list = ();
    my $hdlr2 = new MyFileHandlers;
#    print "Reading file $offlist\n";
    $hdlr2->MFHreadfile($offlist);
    @off_list = $hdlr2->MFHfile;

    my %order_list;
    my $instance;
    my $order = 0;
    my @sorted;
    my $FHO;
    my $out_file;
    my $file_name;
    my $match;
    foreach my $line (@file_bare_bones){
	if($line !~ /\.D\(1\'/){
	    if($line =~ /\s+$ExtractVars::ALLDFF\s+/) {
		$match = $';#'
		    $match =~ s/\s+//g;
	#	print "match = $match\n";
		if($match =~ /($ExtractVars::INSTANCE_NAME).*\(/){
		    if(!defined($1)) {die "1.error in line $line";}
#		    print "match = $1 --- \n";
		    $order_list{$1} = $order;
		    $order = $order + 1;
		   
		}
		
	    }
	}
    }
#    print "order = $order\n";
    @sorted = sort {$order_list{$a} <=> $order_list{$b}} keys %order_list;
    $file_name = "Set.h";
    $out_file = $fileHnd->MFHconcatfile($destdir,$file_name);
    print "Creating file $out_file\n";
    open($FHO,">$out_file") || die "Cannot open $out_file";
    foreach my $val (@sorted){
	print $FHO "$val $order_list{$val}\n";
    }

    foreach my $line (@off_list){
	if($line =~ /($ExtractVars::INSTANCE_NAME)/) {
	    if(!defined($1)) {die "2.error in line $line";}
	    delete $order_list{$1}
	}
    }
    @sorted = sort {$order_list{$a} <=> $order_list{$b}} keys %order_list;
    $file_name = "careSet.h";
    $out_file = $fileHnd->MFHconcatfile($destdir,$file_name);
    print "Creating file $out_file\n";
    open($FHO,">$out_file") || die "Cannot open $out_file";
    foreach my $val (@sorted){
	print $FHO "$val $order_list{$val}\n";
    }
}


sub getSubstrings{
    my $string = shift;
    my $token;
    my $raw_state;
    my $sta;
    my @state;
    if($string =~ /(CONSTRAINT\()/){
	$token = $1;
	$raw_state = substr($string,11);
    }elsif($string =~ /(TARGET\()/){
	$token = $1;
	$raw_state = substr($string,7);
    }	
    chop $raw_state; # remove )
    #push @state, $token;
    $sta = $token;
    my $length =  length($raw_state);
    #print "length = $length raw= $raw_state\n";
    for(my $i=0; $i < ($length); $i = $i + 1){
	if($i == ($length - 1)) {
#	    push @state, substr($raw_state,$i,1) . ")";
	    $sta .= substr($raw_state,$i,1) . ")";
	}else{
#	   push @state, substr($raw_state,$i,1) . ","; 
	    $sta .= substr($raw_state,$i,1) . ","; 
       }
    }
#    print "State=";
#    foreach my $val (@state){
#	print "$val,";
#    }
#    print "\n";
    return $sta;
}


#
#
#
sub countDutFFs{

    my ($ptr) = @_;
    my @file_bare_bones = @$ptr;
    my $count = 0;
    my $i;
    
    for($i = 0; $i <= $#file_bare_bones; $i = $i + 1){
        if($file_bare_bones[$i] =~ /$ExtractVars::ALLDFF/){
	    $count = $count + 1;
	}
    }
    return $count;
}

1;
