#!/usr/bin/env perl

#CVS: $Id: genTB.pl,v 1.3 2009/03/17 21:16:24 depaulfm Exp $

# LICENSE AGREEMENT
#
#  FileName                   [genTB.pl]
#
#  PackageName                [BackSpace]
#
#  Synopsis                   [Use this file for simulation only]
#
#  Author                     [FLAVIO M DE PAULA]
#
#  Copyright                  [Copyright (C) 2009 University of British Columbia]
# 
#  This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

use 5.008;
use warnings;
use Getopt::Long;
use Pod::Usage;
use File::Basename;
use Switch;

use lib "$ENV{'BS_PERL_MODULES'}";

use MyFileHandlers;

#
# Process Input Params
#
my $help           = 0;      # option variable with default value (false)
my $indir          = '';     # option variable for input directory where srcs reside
my $oudir          = '';     # option variable for output directory where target reside
my $inFile         = '';     # option variable containing the extension of the files to be parsed
my $hash_length    = 0; # used for hashing states
my $offlist_length = 0; # used for hashing states

GetOptions ('help|?'  =>  \$help, 
	    'indir=s' => \$indir, 
	    'oudir=s' => \$oudir,
	    'inFile=s' => \$inFile,
	    'hash_length=i'   => \$hash_length,
            'offlist_length=i'   => \$offlist_length);
pod2usage(-exitval => 0, -verbose => 2) if $help;

my $obj = new MyFileHandlers;
my @file_bare_bones = ();
my $destdir = '';

#
# Dictionary 
#
my $ALLDFF        = 'FD1|FD2|FD4';
my $INPUT         = 'input';
my $OUTPUT        = 'output';
my $PORT          = 'input|output';
my $INSTANCE_NAME = '[a-zA-Z0-9_]+';
my $RANGE         = '\[.*\]';

#
# Create absolute path and check that it exists
#
my $currdir = $obj->MFHcurdir;
my $srcdir = $obj->MFHconcatdir($currdir,$indir);

#
# Create a list of files to parsed and modified
#
#$obj->MFHfind($srcdir,'s[0-9]+\.bench');
$obj->MFHfind($srcdir,$inFile);
my @files = $obj->MFHlistoffiles;

# Only read one file
$obj->MFHreadfile($files[0]);
@file_bare_bones = $obj->MFHfile;


$destdir = $obj->MFHconcatdir($currdir,$oudir);
generateTB($destdir,$inFile);


#########################################################
# SUBROUTINES
#########################################################
#
# Semi-automatic generation of a testbench 
#
sub generateTB{
    my ($destdir, $file) = @_;
    
    # misc vars
    my $i;
    my $FH;

    # hold variables w/ specific verilog-type
    my @ports = ();
    my @inputs = ();
    my @DFFs = ();

    # complete path to TB
    my $abs_name;

    # used to hold the DUT (module ... endmodule)
    my @file_bare_bones;

    # used w/ querying stdin 
    my $answ;


    #
    # Load DUT into file_bare_bones
    #
    my @module_name = split ('\.', $file);
    $obj->MFHreadfile($file);
    my @bbones = $obj->MFHfile;
    my $started = 0;
    for($i=0; $i<=$#bbones;$i=$i+1){
	if(($bbones[$i] =~ /endmodule/) && ($started)){	
	    push @file_bare_bones, $bbones[$i];
	}else{
	    if(($bbones[$i] =~ /module $module_name[0]/) || ($started)){
		$started = 1;
		push @file_bare_bones, $bbones[$i];
	    }
	}
    }

    my @date = localtime(time);

    #
    # Gather DUT ports and latches
    #
    @ports = &getPorts(\@file_bare_bones);
    @DFFs = &getDUTDFFs(\@file_bare_bones);
    @inputs = &getInputs(\@file_bare_bones);
   # foreach $in (@inputs){
#	print "Found $in\n";
 #   }
  #  exit();

    # These variables are used for controlling the simulation/breakpoint
    #
    my $nlatches = $#DFFs + 1;

    #
    # Define Verilog header/footer
    #
    my $header = "/*##########
#  This is a testbench file automatically generated by $0 to accompany file $file
#
#  Dependency: 1. 
#              2.
#
#  File created at " . ($date[4] +1) ."/". $date[3] ."/". ($date[5] + 1900). "
###########
*/

`timescale 1ns/1ns

/*#####
BackSpace: Include your own definitions here
######
*/
`define CLK_HALF_PERIOD 10

`define NLATCHES $nlatches
`define NHASH    $hash_length
`define OFFLIST  $offlist_length


/*#####
BackSpace: Include your own include files here
######
*/
`include \"UBC_ISCAS89_sim.v\"



module top; // $module_name[0] TestBench\n


/*#####
BackSpace: Include your parameters definitions here
######
*/

";

    my $footer = " 

endmodule
";

    #
    # Write out file
    # 
    $abs_name = $obj->MFHconcatfile($destdir,$module_name[0]) . "_TB.v";
    print "Writing file $abs_name \n";
    open($FH,">$abs_name") || die "Error: can not open file $abs_name";

    print $FH $header;

    #
    # DUT inputs are type-reg in TB
    # 
    print $FH  "


/*#####
BackSpace: Include your wire/reg definitions here
######
*/
";
    for($i = 0; $i <= $#inputs; $i = $i +1){
	print $FH  "reg " . $inputs[$i] . ";\n";
    }
    
    print $FH  "reg clk;\nreg rst;\n";
    print $FH  "\n\n";
    print $FH  "


/*#####
BackSpace: Data Structures
######
*/
   reg [`NLATCHES-1:0] RESET;
   reg [`NLATCHES-1:0] CONSTRAINT;
   reg [`NLATCHES-1:0] TARGET;
   reg [`NLATCHES-1:0] BREAKPOINT [0:0];
//   reg [`NHASH-1:0]    HASH_VALUE;


/*#####
BackSpace: Clock generator
######
*/
initial
  clk = 'b0;
always begin
   #`CLK_HALF_PERIOD clk = ~clk;
end

/*#####
BackSpace: Reset generator (2 clks on reset)
######
*/ 
initial begin 
  rst = 1'b1;
  #55 rst = 1'b0;
end

/*#####
BackSpace: Read from command-line the seed for pseudo-random stimuli
######
*/
integer seed;
integer status;
initial   status =\$value\$plusargs(\"seed=\%d\",seed);

/*#####
BackSpace: Read from command-line the number of simulation cycles to run
######
*/integer ncycles; 
integer counter;
initial counter = 32\'d0;
initial status =\$value\$plusargs(\"ncycles=\%d\",ncycles);

/*#####
BackSpace: Read from command-line a 1 or 0 to enable/disable reading
 the file 'breakpoint'
######
*/
reg read_bp;
initial 
  begin
     read_bp = 1'b0;
     status =\$value\$plusargs(\"read_bp=\%b\",read_bp);
     if(read_bp === 1'b1)
       \$readmemb(\"breakpoint\",BREAKPOINT);
     //\$display(\"BREAKPOINT:\%b\",BREAKPOINT[0]);
  end

/*#####
BackSpace: Read from command-line the 'timeout' number of simulation cycles 
######
*/
integer timeout; 
initial status =\$value\$plusargs(\"timeout=\%d\",timeout);

/*#####
BackSpace: Open output file 
######
*/
   integer FH, FHhash;
   initial FH = \$fopen(\"states.log\");
   //initial FHhash = \$fopen(\"hash_value.log\");

/*#####
BackSpace: Print the chosen parameters values
######
*/
initial
begin
   \$display(\"##------------##\");
   \$display(\"simv parameters:\");
   \$display(\"##------------##\");
   \$display(\"seed    = \%0d\nncycles = \%0d\ntimeout = \%0d\nread_bp = \%b\",seed,ncycles,timeout,read_bp);
   \$display(\"---------------------------\"\);
end

/*#####
BackSpace: TB fuctions/tasks
######
*/
   //
   // Definition of output_states
   //
   task output_states;
      begin
	 \$fwrite(FH,\"CONSTRAINT(\%b\",CONSTRAINT);
	 \$fdisplay(FH,\")\");
	 \$fwrite(FH,\"TARGET(\%b\",TARGET);
	 \$fdisplay(FH,\")\");
	 \$fdisplay(FH);
	// \$fclose(FH);
      end
   endtask // output_states


   //
   // Definition of get_state
   //
   function [`NLATCHES-1:0] get_state;
      input dummy;
      begin
	 get_state = {
";

    #
    # List 
    # 
    for($i = 0; $i <= $#DFFs; $i = $i +1){
	if ($i < $#DFFs) {
	    print $FH "DUT\.$DFFs[$i],\n";
	}else{
	    print $FH "DUT\.$DFFs[$i]\n
};\n
      end
   endfunction // get_state
";
	}
    }



print $FH "
   //
   // Definition load_hash
   //

/*####
BackSpace: Read from file and load hash value into DUT inputs
#####*/
//  reg [0:0] hash_mem [0:NNNN]; // NNNN = HASH*NLATCHES
//   
//always@(negedge clk)
//  if(rst)  begin
//     \$display(\"[\%0t] Set Hash Function\",\$time);
//     \$readmemb(\"hash_mem.dat\",hash_mem,0,NNNN);
//  end
//  else begin
//`include \"load_hash.v\"
//  end


";


print $FH "
   //
   // BackSpace: Main:: Control simulation
   //
   always@(negedge clk)
     if(rst == 1'b1)
       begin
	  counter = counter + 1;
	  // Captures reset state
	  RESET = get_state(rst);
       end
    else
     begin
	if(counter <= 4)
	  CONSTRAINT = RESET;
	else
	  CONSTRAINT = TARGET;
	TARGET       = get_state(rst);
	//HASH_VALUE   = get_hash_value(rst);

	
	//
	// Checks for termination conditions
	//
	if(read_bp == 1'b0) 
	  begin
	     if(counter >= ncycles)
	       begin
		  output_states;
		  //output_hash_value;
		  \$display(\"------------------------------------\");
		  \$display(\"[\%0t] Simulated \%0d cycles .... finishing\",\$time,counter);
		  #1 \$finish;
	       end
	  end 
	else // running for ncycles
	  begin
	     if(counter > timeout)
	       begin
		  output_states;
		  //output_hash_value;
		  \$display(\"------------------------------------\");
		  \$display(\"[\%0t] Timed out after \%0d cycles ....\",\$time,counter);		  
		  #1 \$finish;
	       end
	     if((RESET[(`NLATCHES-`OFFLIST)-1:`NHASH] === BREAKPOINT[0][`NLATCHES-1:0]))
	       begin
		  \$display(\"------------------------------------\");
		  \$display(\"[\%0t] Backspaced to the RESET state.... exiting\",\$time);
		  #1 \$finish;
	       end 
	     else // simulating till it hits breakpoint
	       if(TARGET[(`NLATCHES-`OFFLIST-1):`NHASH] === BREAKPOINT[0][`NLATCHES-1:0])
		 begin
		    output_states;
		    //output_hash_value;
		    \$display(\"------------------------------------\");
		    \$display(\"[\%0t] Found breakpoint state.... finishing\",\$time);
		    #1 \$finish;
		 end 
	  end // else: !if(read_bp == 1'b0)
	counter = counter + 1;
     end // else: !if(rst === 1'b1)
";


    print $FH  "
/*#####
BackSpace: Include your Test Stimuli
######
*/


/*#####
BackSpace: Instantiate DUT
######
*/
$module_name[0]  DUT(";
    #
    # Identify how clock and reset are called in the DUT and hooked them to the TB's reset and clock
    # The other ports are connected with exact names
    #
    for($i = 0; $i <= $#ports; $i = $i +1){
#	print "printing $ports[$i]\n";
       if($ports[$i] =~ /clk|clock/i) {
	   print $FH  ".$ports[$i] (clk)";
	   print "WARNING: Assuming $ports[$i] to be the design's CLOCK signal: connecting TB's clk to $ports[$i]\n";
       }else{ 
	   if($ports[$i] =~ /rst|reset/i) {
	       print "Is  $ports[$i] a RESET signal(y/n)?";
	        read STDIN, $answ, 2; 
	       if($answ =~ /y/){ 
		   print $FH  ".$ports[$i] (rst)";
		   print "WARNING: Assumed $ports[$i] to be the design's RESET signal: connecting TB's rst to $ports[$i]\n";
	       } else{
		   print $FH  ".$ports[$i] ($ports[$i])";
		   print "WARNING: Connecting $ports[$i] to TB's $ports[$i] instead\n";
	       }
	   }else{
	       print $FH  ".$ports[$i] ($ports[$i])";
	   }
       }
       if ($i < $#ports){print $FH ",\n";}
    }
    print $FH ");\n";

 
    print $FH $footer;

    close($FH);
} ##> end of genTB

#
#
#
sub getInputs{

    my ($ptr) = @_;
    my @file_bare_bones = @$ptr;
    my @list;
    my $i;
    my $allinp = '';
    my $matched;

    my @temp = &prepInputs(\@file_bare_bones);

#    foreach $in (@temp){
#	print "Found $in";
#    }
#    exit();


    for($i = 0; $i <= $#temp; $i = $i + 1){
	if(($temp[$i] =~ /$INPUT(.*)/)) {
	    $matched = $1; # infix from input to ;
	    if ($temp[$i] =~ /\;$/) {
		chop $matched; #remove ';'
		$matched =~ s/\s+//g;
		if ($matched !~ ',') { $matched .= ','}; # if input is singleton
		$allinp .= $matched;
	    }
	}elsif(($temp[$i] !~ /;$/)){ #assumes it ends w/ ','
	    chop $temp[$i];
	    $temp[$i] =~ s/\s+//g;
	    $allinp .= $temp[$i];
	}elsif(($temp[$i] =~ /;$/)){
	    chop $temp[$i];
	    chop $temp[$i]; #remove ';'
	    $temp[$i] =~ s/\s+//g;
	    $allinp .= $temp[$i];
	}
    }
    @list = split(',',$allinp);
#   foreach $in (@list){
#	print "$in\n";
#    }
#    exit();
    return @list;
}

#
#
#
sub  getPorts{

    my ($ptr) = @_;
    my @file_bare_bones = @$ptr;
    my @list;
    my $i;

    my $allinp = '';
    my $matched;


   my @temp = &prepPortlist(\@file_bare_bones);

#   foreach $in (@temp){
#       print "Found $in";
#   }
#    exit();


    for($i = 0; $i <= $#temp; $i = $i + 1){
	if($temp[$i] =~ /($PORT)(.*)/) {
	    $matched = $2; # infix from input to ;
	    if ($temp[$i] =~ /;$/) {
		chop $matched; #remove ';'
		$matched =~ s/\s+//g;
		if ($matched !~ ',$') { $matched .= ','}; #
		$allinp .= $matched;
	    }elsif ($temp[$i] =~ /,$/) {
		$matched =~ s/\s+//g;
		$allinp .= $matched;
	    }
	}elsif(($temp[$i] !~ /;$/)){ #assumes it ends w/ ','
	    chop $temp[$i];
	    $temp[$i] =~ s/\s+//g;
	    $allinp .= $temp[$i];
	}elsif(($temp[$i] =~ /;$/)){
	    chop $temp[$i];
	    chop $temp[$i]; #remove ';'
	    $temp[$i] =~ s/\s+//g;
	    $allinp .= $temp[$i];
	}
    }
    @list = split(',',$allinp);
#   foreach $in (@list){
#       print "$in\n";
#   }
#    exit();

   return @list;
} ###> end newprocessPORTLIST


#
#
#
sub prepInputs{

    my ($ptr) = @_;
    my @file_bare_bones = @$ptr;
    my @list;
    my $i;
    my $allinp = '';
    my $marker = -1;
    for($i = 0; $i <= $#file_bare_bones; $i = $i + 1){
	if($file_bare_bones[$i] =~ /$INPUT(.*)/){
	    $marker = $i;
	    if ($file_bare_bones[$i] =~ /\;$/) {
		push @list, $file_bare_bones[$i]; 
		$marker = -1; # goto next line
	    }elsif(($file_bare_bones[$i] =~ /\,$/) && ($marker > 0) && $file_bare_bones[$marker] =~ /$INPUT(.*)/){
		push @list, $file_bare_bones[$i];
	    }
	}elsif(($file_bare_bones[$i] =~ /\,$/) && ($marker > 0) && $file_bare_bones[$marker] =~ /$INPUT(.*)/){
	    push @list, $file_bare_bones[$i];
	}elsif(($file_bare_bones[$i] =~ /\;$/) && ($marker > 0) && $file_bare_bones[$marker] =~ /$INPUT(.*)/){
	    push @list, $file_bare_bones[$i];
	    $marker = -1;# goto next line
	}
    }
    return @list;
}

#
#
#
sub prepPortlist{

    my ($ptr) = @_;
    my @file_bare_bones = @$ptr;
    my @list;
    my $i;
    my $allinp = '';
    my $marker = -1;
    my $no_range = '';
    for($i = 0; $i <= $#file_bare_bones; $i = $i + 1){
	if($file_bare_bones[$i] =~ /$PORT(.*)/){
	    $marker = $i;
	    if ($file_bare_bones[$i] =~ /\;$/) {
		$no_range = $file_bare_bones[$i];
		$no_range =~ s/$RANGE//g;
		push @list, $no_range;#$file_bare_bones[$i]; 
		$marker = -1; # goto next line
	    }elsif(($file_bare_bones[$i] =~ /\,$/) && ($marker > 0) && $file_bare_bones[$marker] =~ /$PORT(.*)/){
		$no_range = $file_bare_bones[$i];
		$no_range =~ s/$RANGE//g;
		push @list, $no_range;#$file_bare_bones[$i];
	    }
	}elsif(($file_bare_bones[$i] =~ /\,$/) && ($marker > 0) && $file_bare_bones[$marker] =~ /$PORT(.*)/){
	    $no_range = $file_bare_bones[$i];
	    $no_range =~ s/$RANGE//g;
	    push @list, $no_range;#$file_bare_bones[$i];
	}elsif(($file_bare_bones[$i] =~ /\;$/) && ($marker > 0) && $file_bare_bones[$marker] =~ /$PORT(.*)/){
	    $no_range = $file_bare_bones[$i];
	    $no_range =~ s/$RANGE//g;
	    push @list, $no_range;#$file_bare_bones[$i];
	    $marker = -1;# goto next line
	}
    }
	    
    return @list;
}


#
# Handles only the type of flip-flops defined w/ the pre-defined technology library delivered w/ this package
#
sub getDUTDFFs{

    my ($ptr) = @_;
    my @file_bare_bones = @$ptr;
    my @list;
    my $i;
    my $pattern = '$ALLDFF.\s+($INSTANCE_NAME)';
    for($i = 0; $i <= $#file_bare_bones; $i = $i + 1){
	if($file_bare_bones[$i] =~ /$ALLDFF/){
	    if($file_bare_bones[$i] =~ /($INSTANCE_NAME)\s*\(/) {
		#print "line is $file_bare_bones[$i] --$+--\n";
		push @list, $+ . ".flipflop";
	    }
	}
    }
    return @list;
}


