/*
# CVS: $Id: ext_ram.v,v 1.1 2009/03/30 19:42:31 depaulfm Exp $

# LICENSE AGREEMENT
#
#  FileName                   [ext_ram.v]
#
#  PackageName                [BackSpace]
#
#  Synopsis                   [Bus Functional Model based on OpenCore's OpenRisc]
#                             [This file is provided as is. I make no claims    ]
#                             [about its correctness. Use it at your own risk   ]
#
#  Author                     [FLAVIO M DE PAULA]
#
#  Copyright                  [Copyright (C) 2009 University of British Columbia]
# 
#  This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
module   ext_ram(/*AUTOARG*/
   // Outputs
   dout,
   // Inputs
   clk, we, en, addr, din
   );

   //
   // Parameters
   //
   parameter aw = 12;
   parameter number_of_bytes = 2097152; //2MBytes 

   //
   // I/O
   //
   input     clk;
   input we, en;
   input [aw-1:0] addr;
   input [7:0] din;
   output [7:0] dout;
   
// synopsys translate_off   

   //
   // Memory 
   // 
   reg [7:0] 	mem [number_of_bytes-1:0];


   integer 	indx;

`ifdef LOAD_IMAGE
`else
   //
   // initialize mem w/ 0s;
   //
   initial begin
      for(indx=0;indx<number_of_bytes; indx = indx + 1)
	mem[indx] = 32'd0;
   end
`endif
   
   always@(posedge clk) 
     if (we)
       mem[addr] <= din;
   
   assign dout = (en) ? mem[addr] : dout;

// synopsys translate_on

endmodule // onchip_ram
