/*
# CVS: $Id: ext_ram_top.v,v 1.1 2009/03/30 19:42:31 depaulfm Exp $

# LICENSE AGREEMENT
#
#  FileName                   [ext_ram_top.v]
#
#  PackageName                [BackSpace]
#
#  Synopsis                   [Bus Functional Model based on OpenCore's OpenRisc]
#                             [This file is provided as is. I make no claims    ]
#                             [about its correctness. Use it at your own risk   ]
#
#  Author                     [FLAVIO M DE PAULA]
#
#  Copyright                  [Copyright (C) 2009 University of British Columbia]
# 
#  This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
module ext_ram_top (
		       ext_clk_i, ext_rst_i,
		       ext_dat_i, ext_dat_o, ext_adr_i, ext_sel_i, ext_we_i, ext_cyc_i,
		       ext_stb_i, ext_ack_o, ext_err_o
		       );
   //
   // Parameters
   //
   parameter aw = 21; // to be able to access 4GB+ of the vmlinux file //12;
   //
   // I/O Ports
   //
   input     ext_clk_i;
   input     ext_rst_i;
   //
   // slave i/f
   //
   input [31:0] ext_dat_i;
   output [31:0] ext_dat_o;
   input [31:0]  ext_adr_i;
   input [3:0] 	 ext_sel_i;
   input 	 ext_we_i;
   input 	 ext_cyc_i;
   input 	 ext_stb_i;
   output 	 ext_ack_o;
   output 	 ext_err_o;
   //
   // Internal regs and wires
   //
   wire 	 we   = ext_we_i;
   wire [3:0] 	 be_i = ext_sel_i;
   wire [31:0] 	 ext_dat_o;

   
   ext_ram #(aw) block_ram_0 (
			   .addr(ext_adr_i[aw+1:2]),
			   .clk(ext_clk_i),
			   .din(ext_dat_i[7:0]),
			   .dout(ext_dat_o[7:0]),
			   .we(we),
			   .en(be_i[0]));
   ext_ram #(aw) block_ram_1 (
			   .addr(ext_adr_i[aw+1:2]),
			   .clk(ext_clk_i),
			   .din(ext_dat_i[15:8]),
			   .dout(ext_dat_o[15:8]),
			   .we(we),
			   .en(be_i[1]));
   ext_ram #(aw) block_ram_2 (
			   .addr(ext_adr_i[aw+1:2]),
			   .clk(ext_clk_i),
			   .din(ext_dat_i[23:16]),
			   .dout(ext_dat_o[23:16]),
			   .we(we),
			   .en(be_i[2]));
   ext_ram #(aw) block_ram_3 (
			   .addr(ext_adr_i[aw+1:2]),
			   .clk(ext_clk_i),
			   .din(ext_dat_i[31:24]),
			   .dout(ext_dat_o[31:24]),
			   .we(we),
			   .en(be_i[3]));
endmodule