/*
# CVS: $Id: ppc_bfm.v,v 1.1 2009/03/30 19:42:31 depaulfm Exp $

# LICENSE AGREEMENT
#
#  FileName                   [ppc_bfm.v]
#
#  PackageName                [BackSpace]
#
#  Synopsis                   [Bus Functional Model: PowerPC-to-peripheral      ]
#                             [This file is provided as is. I make no claims    ]
#                             [about its correctness. Use it at your own risk   ]
#
#  Author                     [FLAVIO M DE PAULA]
#
#  Copyright                  [Copyright (C) 2009 University of British Columbia]
# 
#  This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
 
module ppc_bfm(/*AUTOARG*/
   // Outputs
   Bus2IP_RdCE, Bus2IP_WrCE, Bus2IP_Data, Bus2IP_BE, sram_address_o,
   sram_data_o, sram_be_o, sram_we_o,
   // Inputs
   clk, reset, IP2Bus_Data, IP2Bus_RdAck, IP2Bus_WrAck, sram_data_i
   );
   
   input clk;
   input reset;
   input [0:63] IP2Bus_Data;
   input IP2Bus_RdAck;
   input IP2Bus_WrAck;
   input [0:31] sram_data_i;
   
   output [0:31] Bus2IP_RdCE;
   output [0:31] Bus2IP_WrCE;
   output [0:63] Bus2IP_Data;	
   output [0:7]  Bus2IP_BE;

   output [0:31] sram_address_o;
   output [0:31] sram_data_o;
   output [0:3]  sram_be_o;
   output 	 sram_we_o;

/*#####
  Encoding of openrisc <-> ppc commands 
#####*/   
//IP2BUS_RdCE commands
parameter  IP2BUS_RAM_SELECT     = 32'd256;
parameter  IP2BUS_RAM_READ       = 32'd32;
parameter  IP2BUS_RAM_WRITE      = 32'd16;
parameter  IP2BUS_RAM_ADDRESS    = 32'd4;
parameter  IP2BUS_RAM_DATA_WRITE = 32'd2;
parameter  NULL_CMD              = 32'd0;
//Bus2IP_WrCE commands
//parameter BUS2IP_RAM_READ        = 32'd1;
//parameter BUS2IP_RAM_ACK         = 32'd8;
   parameter BUS2IP_RAM_READ        = 32'b00000000000000000000000000000001;
   parameter BUS2IP_RAM_ACK         = 32'b00000000000000000000000000001000;
 
// ack from RAM   
parameter RAM_ACK                = 64'd1;
   

   reg [0:31] Bus2IP_RdCE;
   reg [0:31] Bus2IP_WrCE;
   reg [0:63] Bus2IP_Data;	
   reg [0:7]  Bus2IP_BE;

   reg [0:31] 	 sram_address;
   reg [0:31] sram_data;
   reg [0:3]  sram_be;
   reg 	      sram_we;

   wire [0:31] sram_address_o = sram_address;
   wire [0:31] sram_data_o    = sram_data;
   wire [0:3]  sram_be_o      = sram_be;
   wire        sram_we_o      = sram_we;

/* -----\/----- EXCLUDED -----\/-----
   always@(/-*AS*-/Bus2IP_Data or IP2Bus_Data or sram_address_o
	   or sram_data_i)
     begin
	$display("[%0t] i2bd=%b, saddr=%h, sdat=%h, b2ipd=%h",$time,IP2Bus_Data,sram_address_o,sram_data_i,Bus2IP_Data);
     end
 -----/\----- EXCLUDED -----/\----- */
   reg 	       IP2Bus_RdAck_r;

   always@(negedge clk or negedge reset)
     if(reset)
       IP2Bus_RdAck_r = 1'b0;
     else
       IP2Bus_RdAck_r = IP2Bus_RdAck;
   wire        ack_edge = IP2Bus_RdAck & ~IP2Bus_RdAck_r;
   
   initial
     begin
	/* Have to handle events before reset is released and
	 the release of the reset 
	 */
	sram_address = 32'hdeadface;
	sram_data    = 32'hdeadface;
	sram_be      = 4'd0;
	sram_we      = 1'b0;
	Bus2IP_BE    = 8'd0;
	Bus2IP_RdCE = NULL_CMD;
	Bus2IP_WrCE = NULL_CMD;
	Bus2IP_Data = 64'd0;	
 	
	#5;
	
//        while (reset === 1'b1) @(negedge clk);
        while (or1200_tb.DUT.N731 !== 1'b0 || (reset !== 0)) @(posedge clk);
	
	forever
	  begin   
	     @(posedge clk) #1 Bus2IP_RdCE = IP2BUS_RAM_READ;
 	     //$display("[%0t] tx RAM_READ RdCE cmd=%h",$time,Bus2IP_RdCE);
             while (IP2Bus_RdAck !== 1'b1) @(posedge clk);
	     processRead;
	     @(posedge clk) #1 Bus2IP_RdCE = IP2BUS_RAM_WRITE;
    	     ////$display("[%0t] tx RAM_WRITE cmd",$time);
             while (IP2Bus_RdAck !== 1'b1) @(posedge clk);
	     ////$display("[%0t] rx RdAck w/ cmd=%b",$time,IP2Bus_Data[0]);
	     processWrite;
	  end
     end // initial begin
   
   
   
/*#####
  Tasks to decode and execute commands
#####*/   
   
/*#####
 Read command from openrisc
#####*/   
   task processRead;
      begin
	 if(IP2Bus_Data[0] === 1'b1)  //READ
	   begin
	      /* 2-cycle break */
	      ppc_bfm.Bus2IP_RdCE = NULL_CMD; repeat (2)@(posedge clk) ;

	       /* Request address */
	      @(posedge clk) #1 ppc_bfm.Bus2IP_RdCE = IP2BUS_RAM_ADDRESS; //$display("[%0t] tx RAM_ADDRESS RdCE cmd",$time);
              while (IP2Bus_RdAck !== 1'b1) @(posedge clk);
	      sram_address = IP2Bus_Data[0:31];                        //$display("[%0t] rx RdAck w/ address=%h",$time,IP2Bus_Data[0:31],IP2Bus_Data[32:63]);
	      /* 2-cycle break */
	      ppc_bfm.Bus2IP_RdCE = NULL_CMD; repeat (2)@(posedge clk) ;

	      /* Request byte-enable */
	      @(posedge clk) #1 Bus2IP_RdCE = IP2BUS_RAM_SELECT;        //  $display("[%0t] tx RAM_SELECT RdCE cmd",$time);
              while (IP2Bus_RdAck !== 1'b1) @(posedge clk);
	      sram_be = IP2Bus_Data[0:3];                	       //$display("[%0t] rx RdAck w/ be=%h",$time,sram_be);
	      /* 2-cycle break */
	      ppc_bfm.Bus2IP_RdCE = NULL_CMD; repeat (2)@(posedge clk) ;

	      /* Read from memory and relay to openrisc */
	      @(posedge clk) begin
		 #1;
		 sram_we = 1'b0;
		 Bus2IP_Data[0:31] = sram_data_i;
		 Bus2IP_BE = {sram_be,sram_be};
		 Bus2IP_WrCE = BUS2IP_RAM_READ;
	      end                                                      //$display("[%0t] rd from mem=%h (%h)",$time,sram_data_i,sram_be);
              while (IP2Bus_WrAck !== 1'b1) @(posedge clk);
	      /* 2-cycle break */
	      ppc_bfm.Bus2IP_WrCE = NULL_CMD; repeat (2)@(posedge clk) ;

	      /* Send Ack from memory to openrisc  */
	       repeat (1) @(posedge clk) begin
		  #1;
		  Bus2IP_Data[0] = RAM_ACK;
		  Bus2IP_BE = 8'hFF;
		 Bus2IP_WrCE = BUS2IP_RAM_ACK;
	      end
              while (IP2Bus_WrAck !== 1'b1) @(posedge clk);
	      /* 2-cycle break */
	      ppc_bfm.Bus2IP_WrCE = NULL_CMD; repeat (2)@(posedge clk) ;

	      /* Clear Ack from memory to openrisc  */
	      repeat (2) @(posedge clk) begin
		 #1;
		 Bus2IP_Data[0] = ~RAM_ACK;
		 Bus2IP_WrCE = BUS2IP_RAM_ACK;
	      end
              while (IP2Bus_WrAck !== 1'b1) @(posedge clk);
	      /* 2-cycle break */
	      ppc_bfm.Bus2IP_WrCE = NULL_CMD; repeat (2)@(posedge clk) ;
	   end
	 else 
	   begin
	      //move on
	      /* 2-cycle break */
	      ppc_bfm.Bus2IP_WrCE = NULL_CMD; repeat (2)@(posedge clk) ;
	   end
      end
   endtask // processRead
   
 
/*#####
  Write command from openrisc
#####*/   
   task processWrite;
      begin
	 if(IP2Bus_Data[0] === 1'b1)  //WRITE
	   begin
////	      $display("[%0t] First Write instruction... exiting",$time);
////	      $fflush(0);
////	      #10000 $finish;
	      
	      /* 2-cycle break */
	      ppc_bfm.Bus2IP_RdCE = NULL_CMD; repeat (2)@(posedge clk) ;

	      /* Request Data to be written */
	      @(posedge clk) #1 Bus2IP_RdCE = IP2BUS_RAM_DATA_WRITE;
              while (IP2Bus_RdAck !== 1'b1) @(posedge clk);
	      sram_data = IP2Bus_Data[0:31];
	      /* 2-cycle break */
	      ppc_bfm.Bus2IP_RdCE = NULL_CMD; repeat (2)@(posedge clk) ;

	      /* Request Address to be written */
	      @(posedge clk) #1 Bus2IP_RdCE = IP2BUS_RAM_ADDRESS;
              while (IP2Bus_RdAck !== 1'b1) @(posedge clk);
	      sram_address = IP2Bus_Data[0:31];
	      /* 2-cycle break */
	      ppc_bfm.Bus2IP_RdCE = NULL_CMD; repeat (2)@(posedge clk) ;

	      /* Request byte-enable to be written */
	      @(posedge clk) #1 Bus2IP_RdCE = IP2BUS_RAM_SELECT;
              while (IP2Bus_RdAck !== 1'b1) @(posedge clk);
	      sram_be = IP2Bus_Data[0:3];
	      /* 2-cycle break */
	      ppc_bfm.Bus2IP_RdCE = NULL_CMD; repeat (2)@(posedge clk) ;

	      /* Write to memory */
	      @(posedge clk) sram_we = 1'b1;
	      @(posedge clk) sram_we = 1'b0;
	      
	      /* Send Ack from memory to openrisc  */
	       repeat (1) @(posedge clk) begin
		  #1;
		 Bus2IP_Data[0] = RAM_ACK;
		  Bus2IP_BE = 8'hFF;
		 Bus2IP_WrCE = BUS2IP_RAM_ACK;
	      end
              while (IP2Bus_WrAck !== 1'b1) @(posedge clk);
	      /* 2-cycle break */
	      ppc_bfm.Bus2IP_WrCE = NULL_CMD; repeat (2)@(posedge clk) ;

	      /* Clear Ack from memory to openrisc  */
	      repeat (2) @(posedge clk) begin
		  #1;
		 Bus2IP_Data[0] = ~RAM_ACK;
		 Bus2IP_WrCE = BUS2IP_RAM_ACK;
	      end
              while (IP2Bus_WrAck !== 1'b1) @(posedge clk);
	      /* 2-cycle break */
	      ppc_bfm.Bus2IP_WrCE = NULL_CMD; repeat (2)@(posedge clk) ;
	      
	   end 
	 else
	   begin
	      //move on
	      /* 2-cycle break */
	      ppc_bfm.Bus2IP_RdCE = NULL_CMD; repeat (2)@(posedge clk) ;
//	      @(negedge clk) ;
	   end
      end
   endtask // processWrite
   
endmodule // ppc_bfm
